# Copyright(C) 2018-2021  Bruno Chabrier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

# start with:
# set PYTHONPATH=D:\Dropbox\Projets\boomoney
# D:\Dropbox\Projets\boomoney\scripts\bin\woob.exe money -N

import signal
import sys

from threading import Thread, Lock
from io import StringIO
import os
import re
import subprocess
import datetime
from optparse import OptionGroup

import asyncio
from asyncio.subprocess import PIPE
from asyncio import create_subprocess_exec

import shutil

try:
    from colorama import init as init_colorama, Fore, Style
except ModuleNotFoundError:
    HAVE_COLORAMA = False
else:
    HAVE_COLORAMA = True

from woob.capabilities.bank import AccountType
from woob.applications.bank import Appbank
from woob.applications.bank.bank import OfxFormatter
from woob.tools.application.formatters.simple import SimpleFormatter


__all__ = ['AppMoney']


def handler(signum, frame):
    signal.signal(signal.SIGINT, signal.SIG_IGN)

    print('Interrupted', file=sys.__stderr__)
    for t in HistoryThread.allthreads:
        t.terminate()
    sys.exit()


signal.signal(signal.SIGINT, handler)

printMutex = Lock()
numMutex = Lock()
backupMutex = Lock()


class MoneyOfxFormatter(OfxFormatter):

    MANDATORY_FIELDS = tuple(set(OfxFormatter.MANDATORY_FIELDS).union({"type"}) - {"id"})

    def start_format(self, **kwargs):
        self.seen = set()

        # MSMoney only supports CHECKING accounts
        self.original_type = kwargs['account'].type

        # we collect this formatter output because we will need to do some processing to restore the account type
        self.original_outfile = self.outfile
        self.outfile = StringIO()

        kwargs['account'].type = AccountType.CHECKING
        super(MoneyOfxFormatter, self).start_format(**kwargs)
        kwargs['account'].type = self.original_type

    def format_obj(self, obj, alias):
        cat = obj.category
        obj.category = obj.raw
        result = super(MoneyOfxFormatter, self).format_obj(obj, alias)
        obj.category = cat
        return result

    def output(self, formatted):
        if self.outfile != sys.stdout:
            self.outfile.write(formatted + os.linesep)
        else:
            super(MoneyOfxFormatter, self).output(formatted)

    def flush(self):
        super(MoneyOfxFormatter, self).flush()

        # we process the output by restoring the account type
        collected_output = self.outfile.getvalue()
        if hasattr(self, 'original_outfile'):
            # do the test because when killed, flush might be called before original_outfile is called
            self.outfile = self.original_outfile
            self.outfile.write(re.sub('<ACCTTYPE>[^\r]*\r', '<ACCTTYPE>' + str(self.original_type) + '\r', collected_output))


class ListFormatter(SimpleFormatter):
    def output(self, formatted):
        if self.outfile != sys.stdout:
            self.outfile.write(formatted + os.linesep)
        else:
            super(ListFormatter, self).output(formatted)


class HistoryThread(Thread):
    allthreads = []

    def __init__(self, money, accounts):
        Thread.__init__(self)
        self.daemon = True
        HistoryThread.allthreads.append(self)
        self.money = money
        self.accounts = accounts
        self.last_dates = {}
        for a in self.accounts:
            self.last_dates[a] = self.money.config.get(a, 'last_date', default='')

    def terminate(self):
        pass

    def retrieve_history(self, account):
        return self.money.retrieve_history(account)

    def run(self):
        for account in self.accounts:
            last_date = self.money.get_history_from_thread(account, self)
            if not self.money.options.no_import:
                self.last_dates[account] = last_date


class HistoryThreadAsAProcess(HistoryThread):

    def __init__(self, money, accounts):
        super(HistoryThreadAsAProcess, self).__init__(money, accounts)
        self.ofxcontent = ""
        self.stderr = ""
        self.loop = asyncio.new_event_loop()

    def terminate(self):
        self.process.terminate()
        return super(HistoryThreadAsAProcess, self).terminate()

    def retrieve_history(self, account):
        self.ofxcontent[account] = re.sub(r'\r\r\n', r'\n', self.ofxcontent[account])
        return self.ofxcontent[account], self.stderrcontent[account]

    async def _read_stream(self, stream, callback):
        while True:
            line = await stream.readline()
            if line:
                callback(line)
            else:
                break

    async def run_process(self):
        id, backend = self.accounts[0].split("@")

        propagated_options = []
        for o in vars(self.money.options):
            switcher = {
                'backends': False,
                'exclude_backends': False,
                'insecure': True,
                'nss': True,
                'debug': True,
                'quiet': False,
                'verbose': True,
                'logging_file': False,
                'save_responses': False,
                'export_session': False,
                'shell_completion': False,
                'auto_update': False,
                'condition': False,
                'count': True,
                'select': False,
                'formatter': False,
                'no_header': False,
                'no_keys': False,
                'outfile': False,
                'list': False,
                'force': True,
                'accounts': False,
                'until_date': True,
                'no_import': False,
                'display': False
            }
            propagate = switcher.get(o, None)
            if propagate is None:
                self.money.logger.warning("Unhandled option %s." % o)
                propagate = False
            if propagate:
                value = getattr(self.money.options, o)
                o = o.replace("_", "-")
                if value is not None:
                    if value == 0:
                        pass
                    elif value == 1:
                        propagated_options += ["--" + o]
                    else:
                        propagated_options += ["--" + o + "=" + str(value)]

        self.cmd = [
            sys.executable,  # D:\...\woob.exe
            sys.path[0],
            'money',
            '--no-import',
            '--display',
            '--backends=' + backend,
            '--accounts=' + ",".join(self.accounts)
        ] + propagated_options
        self.money.logger.info(" ".join(self.cmd))

        self.process = await create_subprocess_exec(*self.cmd, stdout=PIPE, stderr=PIPE)

        def handle_stdout_line(line):
            line = line.decode("CP1252")
            if not re.search(r"^\([0-9]+/[0-9]+\) ", line) is None:
                account = re.split(r"\s", line)[1]
                self.ofxcontent[account] = self.stdout
                self.stderrcontent[account] = self.stderr
                last_date = self.money.get_history_from_thread(account, self)

                if not self.money.options.no_import:
                    self.last_dates[account] = last_date

                self.stdout = ""
                self.stderr = ""
                self.stdouterr = ""
                self.intransaction = False
                return

            if line.startswith("Hint: There are more results available"):
                self.stderr += line
                self.stdouterr += line
                return

            if line.startswith("OFXHEADER:"):
                self.intransaction = True

            if self.intransaction:
                self.stdout += line
            else:
                self.stderr += line
            self.stdouterr += line

            # print("STDOUT: {}".format(line))

        def handle_stderr_line(line):
            line = line.decode("CP1252")
            self.stderr += line
            self.stdouterr += line
            # print("STDERR: {}".format(line))

        self.stdout = ""
        self.stderr = ""
        self.stdouterr = ""
        self.ofxcontent = dict()
        self.stderrcontent = dict()
        self.intransaction = False

        await asyncio.wait(
                [
                    self._read_stream(self.process.stdout, handle_stdout_line),
                    self._read_stream(self.process.stderr, handle_stderr_line)
                ]
            )

        await self.process.wait()

    def run(self):
        self.loop.run_until_complete(self.run_process())
        if self.stdouterr != "":
            self.money.print(" ".join(self.cmd) + "\n" + self.stdouterr)


class AppMoney(Appbank):
    APPNAME = 'money'
    OLD_APPNAME = 'boomoney'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2018-YEAR Bruno Chabrier'
    DESCRIPTION = "Console application that imports bank accounts into Microsoft Money"
    SHORT_DESCRIPTION = "import bank accounts into Microsoft Money"

    EXTRA_FORMATTERS = {'list': ListFormatter, 'ops_list': MoneyOfxFormatter}
    COMMANDS_FORMATTERS = {'list': 'list', 'history': 'ops_list', 'coming': 'ops_list'}

    def __init__(self):
        super(AppMoney, self).__init__()
        self.importIndex = 0
        application_options = OptionGroup(self._parser, 'Money Options')
        application_options.add_option('-l', '--list', action='store_true', help='list the accounts and balance, without generating any import to MSMoney')
        application_options.add_option('-F', '--force', action='store_true', help='forces the retrieval of transactions (10 maximum), otherwise retrieves only the transactions newer than the previous retrieval date')
        application_options.add_option('-U', '--until-date', help='retrieves until date YYYY-MM-DD max')
        application_options.add_option('-A', '--accounts', help='retrieves only the specified accounts. By default, all accounts are retrieved')
        application_options.add_option('-N', '--no-import', action='store_true', help='no import. Generates the files, but they are not imported in MSMoney. Last import dates are not modified')
        application_options.add_option('-D', '--display', action='store_true', help='displays the generated OFX file')
        self._parser.add_option_group(application_options)
        self.labels = dict()
        self.commands_formatters["select"] = "simple"
        self._backupDone = False

    def str2bool(self, str):
        if str is True:
            return True
        if str is False:
            return False
        if str.upper() == "True".upper():
            return True
        if str.upper() == "False".upper():
            return False
        self.logger.error("Cannot convert '%s' to boolean." % str)
        raise ValueError

    def print(self, *args, **kwargs):
        with printMutex:
            print(*args, **kwargs)

    def write(self, *args):
        with printMutex:
            sys.stdout.write(*args)

    def get_downloads_path(self):
        if not hasattr(self, '_downloadsPath'):
            s = subprocess.check_output('reg query "HKEY_CURRENT_USER\\Software\\Microsoft\\Windows\\CurrentVersion\\Explorer\\User Shell Folders" /v "{374DE290-123F-4565-9164-39C4925E467B}"', encoding='CP850')
            t = re.sub(r'^(.|\r|\n)+REG_EXPAND_SZ\s+([^\n\r]+)(.|\r|\n)*$', r'\2', s)
            self._downloadsPath = os.path.expandvars(t)
        return self._downloadsPath

    def get_money_path(self):
        if not hasattr(self, '_moneyPath'):
            s = subprocess.check_output('reg query HKEY_CLASSES_ROOT\\money\\Shell\\Open\\Command /ve', encoding='CP850')
            t = re.sub(r'^(.|\r|\n)+REG_SZ\s+([^\n\r]+)(.|\r|\n)*$', r'\2', s)
            self._moneyPath = os.path.expandvars(os.path.dirname(t))
        return self._moneyPath

    def get_money_file(self):
        if not hasattr(self, '_moneyFile'):
            s = subprocess.check_output('reg query HKEY_CURRENT_USER\\Software\\Microsoft\\Money\\14.0 /v CurrentFile', encoding='CP850')
            t = re.sub(r'^(.|\r|\n)+REG_SZ\s+([^\n\r]+)(.|\r|\n)*$', r'\2', s)
            self._moneyFile = os.path.expandvars(t)
        return self._moneyFile

    def backup_if_needed(self):
        if not self._backupDone:
            with backupMutex:
                # redo the test in mutual exclusion
                if not (hasattr(self, '_backupDone') and self._backupDone):
                    file = self.get_money_file()
                    filename = os.path.splitext(os.path.basename(file))[0]
                    dir = os.path.dirname(file)
                    self.logger.info(Fore.YELLOW + Style.BRIGHT + "Creating backup of %s..." % file + Style.RESET_ALL)
                    target = os.path.join(dir, filename + datetime.datetime.now().strftime("_%Y_%m_%d_%H%M%S.mny"))
                    shutil.copy2(file, target)
                    self._backupDone = True

    def save_config(self):
        for t in self.threads:
            for a in t.accounts:
                self.config.set(a, 'label', self.config.get(a, 'label', default=''))
                self.config.set(a, 'disabled', self.str2bool(self.config.get(a, 'disabled', default=False)))
                self.config.set(a, 'date_min', self.config.get(a, 'date_min', default=''))
                self.config.set(a, 'last_date', t.last_dates[a])

        self.config.save()

    def get_list(self):
        self.onecmd("select id label number balance")
        self.options.outfile = StringIO()
        self.onecmd("list")
        listContent = self.options.outfile.getvalue()
        self.options.outfile.close()
        self.options.outfile = None

        # find max columns width
        id_maxlength = 0
        label_maxlength = 0
        number_maxlength = 0
        balance_maxlength = 0
        self.ids = []
        labels = []
        balances = []
        self.numbers = []
        for line in listContent.split(os.linesep):
            if not line == "":
                idspec, labelspec, numberspec, balancespec = line.split("\t")
                notusedid, id = idspec.split("=")
                self.ids.append(id)
                id_maxlength = max(id_maxlength, len(id))
                notusedlabel, label = labelspec.split("=")
                bmnlabel = self.config.get(id, 'label', default='')
                if bmnlabel != '' and bmnlabel != label:
                    label = bmnlabel + ' (' + label + ')'
                labels.append(label)
                label_maxlength = max(label_maxlength, len(label))
                notusednumber, number = numberspec.split("=")
                self.numbers.append(number)
                number_maxlength = max(number_maxlength, len(number))
                notusedbalance, balance = balancespec.split("=")
                balances.append(balance)
                balance_maxlength = max(balance_maxlength, len(balance))

                # use the label if not already set
                if self.config.get(id, 'label', default='') == '':
                    self.config.set(id, 'label', label)

        # print columns
        self.print(Style.BRIGHT + "%d accounts:" % len(self.ids))
        sepline = "-".ljust(id_maxlength, "-") + "-" + "-".ljust(number_maxlength, "-") + "-" + "-".ljust(balance_maxlength, "-") + "-" + "-".ljust(label_maxlength, "-") + "-" + "--------"
        self.print(sepline)
        for i in range(len(self.ids)):
            disabled = self.str2bool(self.config.get(self.ids[i], "disabled", default=False))
            self.print(
                self.ids[i].ljust(id_maxlength),
                self.numbers[i].ljust(number_maxlength),
                balances[i].ljust(balance_maxlength),
                labels[i].ljust(label_maxlength),
                "Disabled" if disabled else "")
        self.print(sepline + Style.RESET_ALL)
        self.print()

    def get_history_from_thread(self, account, thread):
        now = datetime.datetime.now().strftime("%Y-%m-%d")
        last_date = self.config.get(account, 'last_date', default='')
        label = self.config.get(account, 'label', default='')

        ofxcontent, stderrcontent = thread.retrieve_history(account)

        if ofxcontent == '':
            self.write(stderrcontent)
            self.importIndex += 1
            self.print(Style.BRIGHT + Fore.RED + "(%i/%i) %s (%s): Got error!" % (
                self.importIndex, self.nb_accounts, account, label) + Style.RESET_ALL)
            return last_date
        else:
            self.handle_ofx_content(account, ofxcontent, stderrcontent)
            return now

    def retrieve_history(self, account):

        date_min = self.config.get(account, 'date_min', default='')
        last_date = self.config.get(account, 'last_date', default='')
        label = self.config.get(account, 'label', default='')

        if self.options.force:
            from_date = date_min
        else:
            if last_date != '':
                from_date = datetime.date.fromisoformat(last_date).isoformat()
            else:
                from_date = ''

        if self.options.count:
            from_date = ''
        if self.options.until_date:
            from_date = self.options.until_date

        self.stderr = StringIO()
        self.stdout = self.stderr
        id, backend = account.split("@")
        module_name, foo = self.woob.backends_config.get_backend(backend)
        moduleHandler = "%s.bat" % os.path.join(os.path.dirname(self.get_money_file()), module_name)
        self.logger.info("Starting history of %s (%s)..." % (account, label))

        MAX_RETRIES = 3
        count = 0
        found = False
        content = ''
        self.error = False
        while count <= MAX_RETRIES and not (found and not self.error):
            self.options.outfile = StringIO()
            self.error = False

            # executing history command
            self.logger.info("select " + " ".join(MoneyOfxFormatter.MANDATORY_FIELDS))
            self.onecmd("select " + " ".join(MoneyOfxFormatter.MANDATORY_FIELDS))
            self.logger.info("history " + account + " " + from_date)
            self.onecmd("history " + account + " " + from_date)
            expected_outputs = 1

            # For CARD accounts, let's also get coming transactions
            # We check number and string until MR!300 is merged. Can take a loooong time...
            regexp = re.compile('\r\n<ACCTTYPE>(' + str(AccountType.CARD) + '|' + list(AccountType._keys)[int(AccountType.CARD)] + ')\r\n')
            if regexp.search(self.options.outfile.getvalue()):
                self.logger.info("coming " + account + " " + from_date)
                self.onecmd("coming " + account + " " + from_date)
                expected_outputs += 1

            historyContent = self.options.outfile.getvalue()
            self.options.outfile.close()
            self.options.outfile = None

            if count > 0:
                self.logger.info("Retrying %s (%s)... %i/%i" % (account, label, count, MAX_RETRIES))
            found = re.findall(r'OFXHEADER:100', historyContent)
            nb_output = len(found)
            if found and nb_output == expected_outputs and not self.error:
                content = historyContent
            count = count + 1

        if content == '':
            # error occurred
            with numMutex:
                self.importIndex = self.importIndex + 1
            self.logger.error("%s (%s): %saborting after %i retries.%s" % (
                account,
                label,
                Fore.RED + Style.BRIGHT,
                MAX_RETRIES,
                Style.RESET_ALL))
            return '', self.stderr.getvalue()

        # postprocessing of the ofx content to match MSMoney expectations
        content = re.sub(r'<BALAMT>Not loaded', r'<BALAMT></BALAMT>', content)
        input = StringIO(content)
        output = StringIO()
        field = {}
        fields = ' '
        transaction = ''
        output_id = 1  # used because we have 2 commands, history and coming
        for line in input:
            if output_id != nb_output:
                # skip trailer of first commands
                if line.startswith('</BANKTRANLIST>'):
                    continue
                if line.startswith('<LEDGERBAL><BALAMT>'):
                    continue
                if line.startswith('<DTASOF>'):
                    continue
                if line.startswith('</LEDGERBAL>'):
                    continue
                if line.startswith('<AVAILBAL><BALAMT>'):
                    continue
                if line.startswith('<DTASOF>20200526</AVAILBAL>'):
                    continue
                if line.startswith('</STMTRS></STMTTRNRS></BANKMSGSRSV1></OFX>'):
                    output_id += 1
                    continue
            if output_id != 1:
                # skip header of last commands
                if line.startswith('OFXHEADER:100'):
                    continue
                if line.startswith('DATA:OFXSGML'):
                    continue
                if line.startswith('VERSION:'):
                    continue
                if line.startswith('SECURITY:NONE'):
                    continue
                if line.startswith('ENCODING:'):
                    continue
                if line.startswith('CHARSET:'):
                    continue
                if line.startswith('COMPRESSION:'):
                    continue
                if line.startswith('OLDFILEUID:'):
                    continue
                if line.startswith('NEWFILEUID:'):
                    continue
                if line.startswith('\r'):
                    continue
                if line.startswith('<OFX><SIGNONMSGSRSV1><SONRS><STATUS><CODE>0<SEVERITY>INFO</STATUS>'):
                    continue
                if line.startswith('<DTSERVER>'):
                    continue
                if line.startswith('<BANKMSGSRSV1><STMTTRNRS><TRNUID>'):
                    continue
                if line.startswith('<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<STMTRS>'):
                    continue
                if line.startswith('<CURDEF>'):
                    continue
                if line.startswith('<BANKID>'):
                    continue
                if line.startswith('<BRANCHID>'):
                    continue
                if line.startswith('<ACCTID>'):
                    continue
                if line.startswith('<ACCTTYPE>'):
                    continue
                if line.startswith('<ACCTKEY>'):
                    continue
                if line.startswith('</BANKACCTFROM>'):
                    continue
                if line.startswith('<BANKTRANLIST>'):
                    continue
                if line.startswith('<DTSTART>'):
                    continue
                if line.startswith('<DTEND>'):
                    continue

            if re.match(r'^OFXHEADER:100', line):
                inTransaction = False
            if re.match(r'^<STMTTRN>', line):
                inTransaction = True
            # MSMoney only supports CHECKING accounts
            if re.match(r'^<ACCTTYPE>', line):
                line = '<ACCTTYPE>CHECKING\r\n'

            if not inTransaction:
                output.write(line)
            else:
                transaction = transaction + line
            if re.match(r'^</STMTTRN>', line):
                # debug: display transaction:
                # print(transaction, file=sys.stderr)
                # if output_id == 2: print(transaction, file=sys.stderr)

                # MSMoney expects CHECKNUM instead of NAME for CHECK transactions
                if "TRNTYPE" in field and field["TRNTYPE"] == "CHECK":
                    if "NAME" in field and str(field["NAME"]).isnumeric():
                        field["CHECKNUM"] = field["NAME"]
                        del field["NAME"]
                        fields = fields.replace(' NAME ', ' CHECKNUM ')

                # go through specific backend process if any
                IGNORE = False
                NEW = None
                origfields = fields
                origfield = field.copy()
                if os.path.exists(moduleHandler):
                    # apply the transformations, in the form
                    # field_NAME=...
                    # field_MEMO=...
                    # field=...

                    cmd = 'cmd /C '
                    for f in field:
                        value = field[f]
                        cmd = cmd + 'set field_%s=%s& ' % (f, value)
                    cmd = cmd + '"' + moduleHandler + '"'

                    self.logger.info(cmd)
                    result = subprocess.Popen(cmd, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
                    (stdout, stderr) = result.communicate()
                    if result.returncode != 0:
                        print(stderr.decode('CP850'), file=sys.stderr)
                        raise subprocess.CalledProcessError(result.returncode, cmd)
                    if stderr != b'' or self.options.verbose is True:
                        if stderr != b'':
                            self.logger.warning(cmd)
                        for f in field:
                            self.print('field_%s=%s' % (f, field[f]))
                        self.print('Output:')
                        self.print(stdout.decode('CP850'), end='')
                        self.print(stderr.decode('CP850'), end='')

                    result = stdout.decode('CP850')

                    for line in re.split(r'[\r\n]+', result):
                        if not line == "":
                            f, value = line.split("=", 1)

                            if f == "IGNORE":
                                IGNORE = True
                            elif f == "NEW":
                                NEW = value
                            elif f.startswith('field_'):
                                f = re.sub(r'^field_', '', f)
                                if value == "":
                                    if f in field:
                                        del field[f]
                                    fields = re.sub(" " + f + " ", " ", fields)
                                else:
                                    field[f] = value
                                    if f not in fields.strip().split(" "):
                                        # MSMoney does not like when CHECKNUM is after MEMO
                                        if f == "CHECKNUM":
                                            fields = fields.replace("MEMO", "CHECKNUM MEMO")
                                        else:
                                            fields = fields + f + " "

                if "DTUSER" in field and "DTPOSTED" in field and not field["DTUSER"] == field["DTPOSTED"]:
                    # the payment date is a deferred payment
                    # MSMoney takes DTPOSTED, which is the payment date
                    # I prefer to have the date of the operation, so I set DTPOSTED
                    # as DTUSER
                    field["DTPOSTED"] = field["DTUSER"]

                if not IGNORE and (from_date == '' or field["DTPOSTED"] >= from_date[0:4] + from_date[5:7] + from_date[8:10]):
                    # dump transaction if not ignored and date matches
                    # (still needed because 'coming "date"' still returns older transactions - bug?
                    #  and anyway, because for CARDs all transactions are listed by 'history "date"' as registered at the payment date)
                    # Also, we keep the transactions listed by 'coming' (output_id == 2)
                    self.dump_transaction(output, fields, field)

                    if NEW is not None:
                        for n in NEW.strip().split(" "):
                            fields = origfields
                            field = origfield.copy()
                            field["FITID"] = origfield["FITID"] + "_" + n
                            for line in re.split(r'[\r\n]+', result):
                                if not line == "":
                                    f, value = line.split("=", 1)

                                    if f.startswith(n + '_field_'):
                                        f = re.sub(r'^.*_field_', '', f)
                                        field[f] = value
                                        if f not in fields.strip().split(" "):
                                            fields = fields + f + " "
                            # dump secondary transaction
                            self.dump_transaction(output, fields, field)

                inTransaction = False
            if inTransaction:
                if re.match(r'^<STMTTRN>', line):
                    field = {}
                    fields = ' '
                    transaction = ''
                else:
                    t = line.split(">", 1)
                    v = re.split(r'[\r\n]', t[1])
                    field[t[0][1:]] = v[0]
                    fields = fields + t[0][1:] + ' '

        ofxcontent = output.getvalue()
        stderrcontent = self.stderr.getvalue()
        input.close()
        output.close()
        self.stderr.close()

        return ofxcontent, stderrcontent

    def dump_transaction(self, output, fields, field):
        output.write("<STMTTRN>\n")
        for f in fields.strip().split(" "):
            value = field[f]
            if f == "NAME":
                if value == "":
                    # MSMoney does not support empty NAME field
                    value = "</NAME>"
                else:
                    # MSMoney does not support NAME field longer than 64
                    value = value[:64]
            output.write("<%s>%s\n" % (f, value))
        output.write("</STMTTRN>\n")

    def handle_ofx_content(self, account, ofxcontent, stderrcontent):

        label = self.config.get(account, 'label', default='')

        if self.options.display:
            self.print(Style.BRIGHT + ofxcontent + Style.RESET_ALL)

        nbTransactions = ofxcontent.count('<STMTTRN>')

        # create ofx file
        fname = re.sub(r'[^\w@\. ]', '_', account + " " + label)
        ofxfile = os.path.join(self.get_downloads_path(), fname + datetime.datetime.now().strftime("_%Y_%m_%d_%H%M%S") + ".ofx")
        with open(ofxfile, "w") as ofx_file:
            ofx_file.write(re.sub(r'\r\n', r'\n', ofxcontent))

        with numMutex:
            self.importIndex = self.importIndex + 1
            index = self.importIndex

        self.write(stderrcontent)

        if not (self.options.no_import or nbTransactions == 0):
            self.backup_if_needed()
        with printMutex:
            if self.options.no_import or nbTransactions == 0:
                if nbTransactions == 0:
                    print(Style.BRIGHT + '(%i/%i) %s (%s) (no transaction).' % (
                        index, self.nb_accounts,
                        account,
                        label
                    ) + Style.RESET_ALL)
                else:
                    print(Fore.GREEN + Style.BRIGHT + '(%i/%i) %s (%s) (%i transaction(s)).' % (
                        index, self.nb_accounts,
                        account,
                        label,
                        nbTransactions
                    ) + Style.RESET_ALL)
            else:
                # import into money
                print(Fore.GREEN + Style.BRIGHT + '(%i/%i) Importing "%s" into MSMoney (%i transaction(s))...' % (
                    index, self.nb_accounts,
                    ofxfile,
                    nbTransactions
                ) + Style.RESET_ALL)
        if not self.options.no_import:
            if nbTransactions > 0:
                subprocess.check_call('"%s" %s' % (
                    os.path.join(self.get_money_path(), "mnyimprt.exe"),
                    ofxfile))

    def main(self, argv):

        if not HAVE_COLORAMA:
            print('Error: Please install the "colorama" python package', file=sys.stderr)
            return 1

        init_colorama()

        self.load_config()

        self._interactive = False

        if self.options.list:
            self.get_list()
            return

        self.threads = set()

        accounts = []
        if not self.options.accounts:
            self.get_list()
            accounts = self.ids
        else:
            accounts = self.options.accounts.split(",")

        # take only enabled accounts
        accounts = list(filter(
            lambda account: self.str2bool(self.config.get(account, "disabled", default=False)) is False,
            accounts))
        self.nb_accounts = len(accounts)

        # accounts.sort(key = lambda x: x.split("@")[1])

        # make a unique list
        backends = list(set(map(lambda x: x.split("@")[1], accounts)))

        # order backends by number of accounts
        backends.sort(key=lambda b: len(list(filter(
                lambda x: x.split("@")[1] == b,
                accounts))), reverse=True)

        self.main_thread = None
        for backend in backends:
            backend_accounts = list(filter(
                lambda x: x.split("@")[1] == backend,
                accounts))

            if backend == backends[0]:
                self.main_thread = HistoryThread(self, backend_accounts)
                self.threads.add(self.main_thread)
            else:
                self.threads.add(HistoryThreadAsAProcess(self, backend_accounts))

        if self.main_thread is not None:
            for t in self.threads:
                if t != self.main_thread:
                    t.start()
            self.main_thread.run()

            for t in self.threads:
                if t != self.main_thread:
                    t.join()

        self.save_config()
        return



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import Capability, BaseObject, StringField, IntField, UserError


__all__ = ['Subtitle', 'CapSubtitle']


class LanguageNotSupported(UserError):
    """
    Raised when the language is not supported
    """

    def __init__(self, msg='language is not supported'):
        super(LanguageNotSupported, self).__init__(msg)


class Subtitle(BaseObject):
    """
    Subtitle object.
    """
    name =      StringField('Name of subtitle')
    ext =       StringField('Extension of file')
    nb_cd =     IntField('Number of cd or files')
    language =  StringField('Language of the subtitle')
    description=StringField('Description of corresponding video')

    def __init__(self, id='', name='', url=None):
        super(Subtitle, self).__init__(id, url)
        self.name = name


class CapSubtitle(Capability):
    """
    Subtitle providers.
    """

    def iter_subtitles(self, pattern):
        """
        Search subtitles and iterate on results.

        :param pattern: pattern to search
        :type pattern: str
        :rtype: iter[:class:`Subtitle`]
        """
        raise NotImplementedError()

    def get_subtitle(self, _id):
        """
        Get a subtitle object from an ID.

        :param _id: ID of subtitle
        :type _id: str
        :rtype: :class:`Subtitle`
        """
        raise NotImplementedError()

    def get_subtitle_file(self, _id):
        """
        Get the content of the subtitle file.

        :param _id: ID of subtitle
        :type _id: str
        :rtype: str
        """
        raise NotImplementedError()



# Copyright(C) 2014-2023 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import http.cookiejar
import requests.cookies


__all__ = ['WoobCookieJar', 'BlockAllCookies']


class WoobCookieJar(requests.cookies.RequestsCookieJar):
    @classmethod
    def from_cookiejar(klass, cj):
        """
        Create a WoobCookieJar from another CookieJar instance.
        """
        return requests.cookies.merge_cookies(klass(), cj)

    def export(self, filename):
        """
        Export all cookies to a file, regardless of expiration, etc.
        """
        cj = requests.cookies.merge_cookies(http.cookiejar.LWPCookieJar(), self)
        cj.save(filename, ignore_discard=True, ignore_expires=True)

    def copy(self):
        """Return an object copy of the cookie jar."""
        new_cj = type(self)()
        if hasattr(self, 'get_policy'):
            new_cj.set_policy(self.get_policy())
        else:
            new_cj.set_policy(self._policy)
        new_cj.update(self)
        return new_cj


WeboobCookieJar = WoobCookieJar


class BlockAllCookies(http.cookiejar.CookiePolicy):
    return_ok = set_ok = domain_return_ok = path_return_ok = lambda self, *args, **kwargs: False
    netscape = True
    rfc2965 = hide_cookie2 = False



# Copyright(C) 2009-2011  Romain Bignon, Christophe Benz
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from contextlib import contextmanager
import datetime
import uuid
from dateutil.relativedelta import relativedelta
from dateutil.parser import parse as parse_date
from decimal import Decimal, InvalidOperation

from woob.exceptions import (
    CaptchaQuestion, DecoupledValidation,
    AppValidationCancelled, AppValidationExpired,
)
from woob.core.bcall import CallErrors
from woob.capabilities.base import empty, find_object
from woob.capabilities.bank import (
    Account, Transaction,
    Transfer, TransferStep, Recipient, AddRecipientStep,
    CapBank, CapTransfer,
    TransferInvalidLabel, TransferInvalidAmount, TransferInvalidDate,
    TransferInvalidEmitter, TransferInvalidRecipient,
)
from woob.capabilities.bank.wealth import CapBankWealth
from woob.capabilities.captcha import exception_to_job
from woob.capabilities.profile import CapProfile
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.captcha import CaptchaMixin
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.tools.misc import to_unicode


__all__ = ['Appbank']


class OfxFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'rdate', 'label', 'raw', 'amount', 'category')
    TYPES_ACCTS = {
        Account.TYPE_CHECKING: 'CHECKING',
        Account.TYPE_SAVINGS: 'SAVINGS',
        Account.TYPE_DEPOSIT: 'DEPOSIT',
        Account.TYPE_LOAN: 'LOAN',
        Account.TYPE_MARKET: 'MARKET',
        Account.TYPE_JOINT: 'JOINT',
        Account.TYPE_CARD: 'CARD',
    }
    TYPES_TRANS = {
        Transaction.TYPE_TRANSFER: 'DIRECTDEP',
        Transaction.TYPE_ORDER: 'PAYMENT',
        Transaction.TYPE_CHECK: 'CHECK',
        Transaction.TYPE_DEPOSIT: 'DEP',
        Transaction.TYPE_PAYBACK: 'OTHER',
        Transaction.TYPE_WITHDRAWAL: 'ATM',
        Transaction.TYPE_CARD: 'POS',
        Transaction.TYPE_INSTANT: 'POS',
        Transaction.TYPE_LOAN_PAYMENT: 'INT',
        Transaction.TYPE_BANK: 'FEE',
    }

    balance = Decimal(0)
    coming = Decimal(0)
    account_type = ''
    seen = set()

    def start_format(self, **kwargs):
        account = kwargs['account']
        self.balance = account.balance
        self.coming = account.coming
        self.account_type = account.type

        self.output('OFXHEADER:100')
        self.output('DATA:OFXSGML')
        self.output('VERSION:102')
        self.output('SECURITY:NONE')
        self.output('ENCODING:UTF-8')
        self.output('CHARSET:UTF-8')
        self.output('COMPRESSION:NONE')
        self.output('OLDFILEUID:NONE')
        self.output('NEWFILEUID:%s\n' % uuid.uuid1())
        self.output('<OFX><SIGNONMSGSRSV1><SONRS><STATUS><CODE>0<SEVERITY>INFO</STATUS>')
        self.output('<DTSERVER>%s113942<LANGUAGE>ENG</SONRS></SIGNONMSGSRSV1>' % datetime.date.today().strftime('%Y%m%d'))

        if self.account_type == Account.TYPE_CARD:
            self.output('<CREDITCARDMSGSRSV1><CCSTMTTRNRS><TRNUID>%s' % uuid.uuid1())
            self.output('<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<CCSTMTRS>')
            self.output('<CURDEF>%s<CCACCTFROM>' % (account.currency or 'EUR'))
            self.output('<BANKID>null')
            self.output('<BRANCHID>null')
            self.output('<ACCTID>%s' % account.id)
            self.output('<ACCTTYPE>%s' % self.account_type)
            self.output('<ACCTKEY>null')
            self.output('</CCACCTFROM>')
        else:
            self.output('<BANKMSGSRSV1><STMTTRNRS><TRNUID>%s' % uuid.uuid1())
            self.output('<STATUS><CODE>0<SEVERITY>INFO</STATUS><CLTCOOKIE>null<STMTRS>')
            self.output('<CURDEF>%s<BANKACCTFROM>' % (account.currency or 'EUR'))
            self.output('<BANKID>null')
            self.output('<BRANCHID>null')
            self.output('<ACCTID>%s' % account.id)
            self.output('<ACCTTYPE>%s' % self.account_type)
            self.output('<ACCTKEY>null')
            self.output('</BANKACCTFROM>')

        self.output('<BANKTRANLIST>')
        self.output('<DTSTART>%s' % datetime.date.today().strftime('%Y%m%d'))
        self.output('<DTEND>%s' % datetime.date.today().strftime('%Y%m%d'))

    def format_obj(self, obj, alias):
        # special case of coming operations with card ID
        result = '<STMTTRN>\n'
        if hasattr(obj, '_coming') and obj._coming and hasattr(obj, 'obj._cardid') and not empty(obj._cardid):
            result += '<TRNTYPE>%s\n' % obj._cardid
        elif obj.type in self.TYPES_TRANS:
            result += '<TRNTYPE>%s\n' % self.TYPES_TRANS[obj.type]
        else:
            result += '<TRNTYPE>%s\n' % ('DEBIT' if obj.amount < 0 else 'CREDIT')

        result += '<DTPOSTED>%s\n' % obj.date.strftime('%Y%m%d')
        if obj.rdate:
            result += '<DTUSER>%s\n' % obj.rdate.strftime('%Y%m%d')
        result += '<TRNAMT>%s\n' % obj.amount
        result += '<FITID>%s\n' % obj.unique_id(self.seen)

        if hasattr(obj, 'label') and not empty(obj.label):
            result += '<NAME>%s\n' % obj.label.replace('&', '&amp;')
        else:
            result += '<NAME>%s\n' % obj.raw.replace('&', '&amp;')
        if obj.category:
            result += '<MEMO>%s\n' % obj.category.replace('&', '&amp;')
        result += '</STMTTRN>'

        return result

    def flush(self):
        self.output('</BANKTRANLIST>')
        self.output('<LEDGERBAL><BALAMT>%s' % self.balance)
        self.output('<DTASOF>%s' % datetime.date.today().strftime('%Y%m%d'))
        self.output('</LEDGERBAL>')

        try:
            self.output('<AVAILBAL><BALAMT>%s' % (self.balance + self.coming))
        except TypeError:
            self.output('<AVAILBAL><BALAMT>%s' % self.balance)
        self.output('<DTASOF>%s</AVAILBAL>' % datetime.date.today().strftime('%Y%m%d'))

        if self.account_type == Account.TYPE_CARD:
            self.output('</CCSTMTRS></CCSTMTTRNRS></CREDITCARDMSGSRSV1></OFX>')
        else:
            self.output('</STMTRS></STMTTRNRS></BANKMSGSRSV1></OFX>')


class QifFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'raw', 'amount')

    def start_format(self, **kwargs):
        self.output('!Type:Bank')

    def format_obj(self, obj, alias):
        result = 'D%s\n' % obj.date.strftime('%d/%m/%y')
        result += 'T%s\n' % obj.amount
        if hasattr(obj, 'category') and not empty(obj.category):
            result += 'N%s\n' % obj.category
        result += 'M%s\n' % obj.raw
        result += '^'
        return result


class PrettyQifFormatter(QifFormatter):
    MANDATORY_FIELDS = ('id', 'date', 'raw', 'amount', 'category')

    def start_format(self, **kwargs):
        self.output('!Type:Bank')

    def format_obj(self, obj, alias):
        if hasattr(obj, 'rdate') and not empty(obj.rdate):
            result = 'D%s\n' % obj.rdate.strftime('%d/%m/%y')
        else:
            result = 'D%s\n' % obj.date.strftime('%d/%m/%y')
        result += 'T%s\n' % obj.amount

        if hasattr(obj, 'category') and not empty(obj.category):
            result += 'N%s\n' % obj.category

        if hasattr(obj, 'label') and not empty(obj.label):
            result += 'M%s\n' % obj.label
        else:
            result += 'M%s\n' % obj.raw

        result += '^'
        return result


class TransactionsFormatter(IFormatter):
    MANDATORY_FIELDS = ('date', 'label', 'amount')
    TYPES = [
        '',
        'Transfer',
        'Order',
        'Check',
        'Deposit',
        'Payback',
        'Withdrawal',
        'Card',
        'Loan',
        'Bank',
        'Cash deposit',
        'Card summary',
        'Deferred card'
    ]

    def start_format(self, **kwargs):
        self.output(' Date         Category     Label                                                  Amount ')
        self.output('------------+------------+---------------------------------------------------+-----------')

    def format_obj(self, obj, alias):
        if hasattr(obj, 'category') and obj.category:
            _type = obj.category
        else:
            try:
                _type = self.TYPES[obj.type]
            except (IndexError, AttributeError):
                _type = ''

        label = obj.label or ''
        if not label and hasattr(obj, 'raw'):
            label = obj.raw or ''
        date = obj.date.strftime('%Y-%m-%d') if not empty(obj.date) else ''
        amount = obj.amount or Decimal('0')
        return ' %s   %s %s %s' % (
            self.colored('%-10s' % date, 'blue'),
            self.colored('%-12s' % _type[:12], 'magenta'),
            self.colored('%-50s' % label[:50], 'yellow'),
            self.colored('%10.2f' % amount, 'green' if amount >= 0 else 'red')
        )


class TransferFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'exec_date', 'account_label', 'recipient_label', 'amount')
    DISPLAYED_FIELDS = ('label', 'account_iban', 'recipient_iban', 'currency')

    def format_obj(self, obj, alias):
        result = '------- Transfer %s -------\n' % obj.fullid
        result += 'Date:       %s\n' % obj.exec_date
        if obj.account_iban:
            result += 'Origin:     %s (%s)\n' % (obj.account_label, obj.account_iban)
        else:
            result += 'Origin:     %s\n' % obj.account_label

        if obj.recipient_iban:
            result += 'Recipient:  %s (%s)\n' % (obj.recipient_label, obj.recipient_iban)
        else:
            result += 'Recipient:  %s\n' % obj.recipient_label

        result += 'Amount:     %.2f %s\n' % (obj.amount, obj.currency or '')
        if obj.label:
            result += 'Label:      %s\n' % obj.label
        return result


class TransferListFormatter(IFormatter):
    def format_obj(self, obj, alias):
        result = [
            'From: %s' % self.colored('%-20s' % obj.account_label, 'red'),
            ' Label: %s\n' % self.colored(obj.label, 'yellow'),
            'To: %s' % self.colored('%-22s' % obj.recipient_label, 'green'),
            ' Amount: %s\n' % self.colored(obj.amount, 'red'),
            'Date: %s' % self.colored(obj.exec_date, 'yellow'),
            ' Status: %s' % self.colored(obj.status, 'yellow'),
            '\n',
        ]
        return ''.join(result)


class InvestmentFormatter(IFormatter):
    MANDATORY_FIELDS = ('label', 'quantity', 'unitvalue')
    DISPLAYED_FIELDS = ('code', 'diff')

    tot_valuation = Decimal(0)
    tot_diff = Decimal(0)

    def start_format(self, **kwargs):
        self.output(' Label                            Code          Quantity     Unit Value   Valuation    diff    ')
        self.output('-------------------------------+--------------+------------+------------+------------+---------')

    def check_emptyness(self, obj):
        if not empty(obj):
            return (obj, '%11.2f')
        return ('---', '%11s')

    def format_obj(self, obj, alias):
        label = obj.label

        if not empty(obj.diff):
            diff = obj.diff
        elif not empty(obj.quantity) and not empty(obj.unitprice):
            diff = obj.valuation - (obj.quantity * obj.unitprice)
        else:
            diff = '---'
            format_diff = '%8s'
        if isinstance(diff, Decimal):
            format_diff = '%8.2f'
            self.tot_diff += diff

        if not empty(obj.quantity):
            quantity = obj.quantity
            format_quantity = '%11.2f'
            if obj.quantity == obj.quantity.to_integral():
                format_quantity = '%11d'
        else:
            format_quantity = '%11s'
            quantity = '---'

        unitvalue, format_unitvalue = self.check_emptyness(obj.unitvalue)
        valuation, format_valuation = self.check_emptyness(obj.valuation)
        if isinstance(valuation, Decimal):
            self.tot_valuation += obj.valuation

        if empty(obj.code) and not empty(obj.description):
            code = obj.description
        else:
            code = obj.code

        return ' %s  %s  %s  %s  %s  %s' % \
            (
                self.colored('%-30s' % label[:30], 'red'),
                self.colored('%-12s' % code[:12], 'yellow') if not empty(code) else ' ' * 12,
                self.colored(format_quantity % quantity, 'yellow'),
                self.colored(format_unitvalue % unitvalue, 'yellow'),
                self.colored(format_valuation % valuation, 'yellow'),
                self.colored(format_diff % diff, 'green' if not isinstance(diff, str) and diff >= 0 else 'red')
            )

    def flush(self):
        self.output('-------------------------------+--------------+------------+------------+------------+---------')
        self.output('                                                                  Total  %s %s'
            % (
                self.colored('%11.2f' % self.tot_valuation, 'yellow'),
                self.colored('%9.2f' % self.tot_diff, 'green' if self.tot_diff >= 0 else 'red')
            )
        )
        self.tot_valuation = Decimal(0)
        self.tot_diff = Decimal(0)


class RecipientListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'label')
    DISPLAYED_FIELDS = ('iban', 'bank_name')

    def start_format(self, **kwargs):
        self.output('Available recipients:')

    def get_title(self, obj):
        details = ' - '.join(filter(None, (obj.iban, obj.bank_name)))
        if details:
            return '%s (%s)' % (obj.label, details)
        return obj.label


class AdvisorListFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'name')

    def start_format(self, **kwargs):
        self.output('   Bank           Name                           Contacts')
        self.output('-----------------+------------------------------+-----------------------------------------')

    def format_obj(self, obj, alias):
        bank = obj.backend
        phones = ""
        contacts = []
        if not empty(obj.phone):
            phones += obj.phone
        if not empty(obj.mobile):
            if phones != "":
                phones += " or %s" % obj.mobile
            else:
                phones += obj.mobile
        if phones:
            contacts.append(phones)

        for attr in ('email', 'agency', 'address'):
            value = getattr(obj, attr)
            if not empty(value):
                contacts.append(value)

        if len(contacts) > 0:
            first_contact = contacts.pop(0)
        else:
            first_contact = ""

        result = "  %s %s %s " % (
            self.colored('%-15s' % bank, 'yellow'),
            self.colored('%-30s' % obj.name, 'red'),
            self.colored("%-30s" % first_contact, 'green')
        )
        for contact in contacts:
            result += "\n %s %s" % ((" ") * 47, self.colored("%-25s" % contact, 'green'))

        return result


class AccountListFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'label', 'balance', 'coming', 'type')

    totals = {}

    def start_format(self, **kwargs):
        self.output('               %s  Account                     Balance    Coming ' % ((' ' * 15) if not self.interactive else ''))
        self.output('------------------------------------------%s+----------+----------' % (('-' * 15) if not self.interactive else ''))

    def format_obj(self, obj, alias):
        if alias is not None:
            id = '%s (%s)' % (self.colored('%3s' % ('#' + alias), 'red', 'bold'),
                              self.colored(obj.backend, 'blue', 'bold'))
            clean = '#%s (%s)' % (alias, obj.backend)
            if len(clean) < 15:
                id += (' ' * (15 - len(clean)))
        else:
            id = self.colored('%30s' % obj.fullid, 'red', 'bold')

        balance = obj.balance or Decimal('0')
        coming = obj.coming or Decimal('0')
        currency = obj.currency or 'EUR'
        result = '%s %s %s  %s' % (
            id,
            self.colored('%-25s' % obj.label[:25], 'yellow' if obj.type != Account.TYPE_LOAN else 'blue'),
            self.colored('%9.2f' % obj.balance, 'green' if balance >= 0 else 'red') if not empty(obj.balance) else ' ' * 9,
            self.colored('%9.2f' % obj.coming, 'green' if coming >= 0 else 'red') if not empty(obj.coming) else ''
        )

        currency_totals = self.totals.setdefault(currency, {})
        currency_totals.setdefault('balance', Decimal(0))
        currency_totals.setdefault('coming', Decimal(0))

        if obj.type != Account.TYPE_LOAN:
            currency_totals['balance'] += balance
            currency_totals['coming'] += coming
        return result

    def flush(self):
        self.output('------------------------------------------%s+----------+----------' % (('-' * 15) if not self.interactive else ''))
        for currency, currency_totals in sorted(self.totals.items(), key=lambda k_v: (k_v[1]['balance'], k_v[1]['coming'], k_v[0])):
            balance = currency_totals['balance']
            coming = currency_totals['coming']

            self.output(
                '%s                              Total (%s)   %s   %s'
                % (
                    (' ' * 15) if not self.interactive else '',
                    currency,
                    self.colored('%8.2f' % balance, 'green' if balance >= 0 else 'red'),
                    self.colored('%8.2f' % coming, 'green' if coming >= 0 else 'red')
                )
            )
        self.totals.clear()


class EmitterListFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'label', 'currency')

    def start_format(self, **kwargs):
        self.output(
            '       %s  Emitter              Currency   Number Type      Number     Balance '
            % (
                (' ' * 15) if not self.interactive else ''
            )
        )
        self.output(
            '----------------------------%s+----------+-------------+-------------+----------+' % (
                ('-' * 15) if not self.interactive else ''
            )
        )

    def format_emitter_number(self, obj):
        account_number = ' '*11
        if obj.number_type != 'unknown' and obj.number:
            account_number = '%s***%s' % (obj.number[:4], obj.number[len(obj.number) - 4:])
        return account_number

    def format_obj(self, obj, alias):
        if alias is not None:
            obj_id = '%s' % self.colored('%3s' % ('#' + alias), 'red', 'bold')
        else:
            obj_id = self.colored('%30s' % obj.fullid, 'red', 'bold')

        balance = ' ' * 9
        if not empty(obj.balance):
            balance = self.colored('%9.2f' % obj.balance, 'green' if obj.balance >= 0 else 'red')

        account_number = self.format_emitter_number(obj)

        return '%s %s %s %s %s %s' % (
            obj_id,
            self.colored('%-25s' % obj.label[:25], 'yellow', 'bold'),
            self.colored('%-10s' % obj.currency, 'green', 'bold'),
            self.colored('%-13s' % obj.number_type, 'blue', 'bold'),
            self.colored('%-11s' % account_number, 'blue', 'bold'),
            balance
        )

    def flush(self):
        self.output(
            '----------------------------%s+----------+-------------+-------------+----------+' % (
                ('-' * 15) if not self.interactive else ''
            )
        )


class Appbank(CaptchaMixin, ReplApplication):
    APPNAME = 'bank'
    OLD_APPNAME = 'boobank'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon, Christophe Benz'
    CAPS = CapBank
    DESCRIPTION = "Console application allowing to list your bank accounts and get their balance, " \
                  "display accounts history and coming bank operations, and transfer money from an account to " \
                  "another (if available)."
    SHORT_DESCRIPTION = "manage bank accounts"
    EXTRA_FORMATTERS = {
        'account_list':    AccountListFormatter,
        'recipient_list':  RecipientListFormatter,
        'transfer':        TransferFormatter,
        'qif':             QifFormatter,
        'pretty_qif':      PrettyQifFormatter,
        'ofx':             OfxFormatter,
        'ops_list':        TransactionsFormatter,
        'investment_list': InvestmentFormatter,
        'advisor_list':    AdvisorListFormatter,
        'transfer_list':   TransferListFormatter,
        'emitter_list':    EmitterListFormatter,
    }
    DEFAULT_FORMATTER = 'table'
    COMMANDS_FORMATTERS = {
        'ls':          'account_list',
        'list':        'account_list',
        'recipients':  'recipient_list',
        'transfer':    'transfer',
        'history':     'ops_list',
        'coming':      'ops_list',
        'transfer_history': 'transfer_list',
        'investment':  'investment_list',
        'advisor':     'advisor_list',
        'emitters':    'emitter_list',
    }
    COLLECTION_OBJECTS = (Account, Transaction, )

    def bcall_error_handler(self, backend, error, backtrace):
        if isinstance(error, TransferStep):
            params = {}
            for field in error.fields:
                v = self.ask(field)
                params[field.id] = v
            #backend.config['accept_transfer'].set(v)
            params['backends'] = backend
            self.start_format()
            for transfer in self.do('transfer', error.transfer, **params):
                self.format(transfer)
        elif isinstance(error, AddRecipientStep):
            params = {}
            params['backends'] = backend
            for field in error.fields:
                v = self.ask(field)
                params[field.id] = v
            try:
                next(iter(self.do('add_recipient', error.recipient, **params)))
            except CallErrors as e:
                self.bcall_errors_handler(e)
        elif isinstance(error, DecoupledValidation):
            if isinstance(error.resource, Recipient):
                func_name = 'add_recipient'
            elif isinstance(error.resource, Transfer):
                func_name = 'transfer'
            else:
                print('Error(%s): The resource should be of type Recipient or Transfer, not "%s"' % (backend.name, type(error.resource)), file=self.stderr)
                return False

            print(error.message)
            params = {
                'backends': backend,
                'resume': True,
            }
            try:
                next(iter(self.do(func_name, error.resource, **params)))
            except CallErrors as e:
                self.bcall_errors_handler(e)
        elif isinstance(error, AppValidationCancelled):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The app validation has been cancelled'), file=self.stderr)
        elif isinstance(error, AppValidationExpired):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The app validation has expired'), file=self.stderr)
        elif isinstance(error, TransferInvalidAmount):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer amount is invalid'), file=self.stderr)
        elif isinstance(error, TransferInvalidLabel):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer label is invalid'), file=self.stderr)
        elif isinstance(error, TransferInvalidEmitter):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer emitter is invalid'), file=self.stderr)
        elif isinstance(error, TransferInvalidRecipient):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer recipient is invalid'), file=self.stderr)
        elif isinstance(error, TransferInvalidDate):
            print('Error(%s): %s' % (backend.name, to_unicode(error) or 'The transfer execution date is invalid'), file=self.stderr)
        elif isinstance(error, CaptchaQuestion):
            if not self.captcha_woob.count_backends():
                print('Error(%s): Site requires solving a CAPTCHA but no CapCaptchaSolver backends were configured' % backend.name, file=self.stderr)
                return False

            print('Info(%s): Encountered CAPTCHA, please wait for its resolution, it can take dozens of seconds' % backend.name, file=self.stderr)
            job = exception_to_job(error)
            self.solve_captcha(job, backend)
            return False
        else:
            return super(Appbank, self).bcall_error_handler(backend, error, backtrace)

    def load_default_backends(self):
        self.load_backends(CapBank, storage=self.create_storage())

    def _complete_account(self, exclude=None):
        if exclude:
            exclude = '%s@%s' % self.parse_id(exclude)

        return [s for s in self._complete_object() if s != exclude]

    def do_list(self, line):
        """
        list [-U]

        List accounts.
        Use -U to disable sorting of results.
        """
        return self.do_ls(line)

    def show_history(self, command, line):
        id, end_date = self.parse_command_args(line, 2, 1)

        account = self.get_object(id, 'get_account', [])
        if not account:
            print('Error: account "%s" not found (Hint: try the command "list")' % id, file=self.stderr)
            return 2

        if end_date is not None:
            try:
                end_date = parse_date(end_date)
            except ValueError:
                print('"%s" is an incorrect date format (for example "%s")' %
                      (end_date, (datetime.date.today() - relativedelta(months=1)).strftime('%Y-%m-%d')), file=self.stderr)
                return 3
            old_count = self.options.count
            self.options.count = None

        self.start_format(account=account)
        for transaction in self.do(command, account, backends=account.backend):
            if end_date is not None and transaction.date < end_date:
                break
            self.format(transaction)

        if end_date is not None:
            self.options.count = old_count

    def complete_history(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_account()

    @defaultcount(10)
    def do_history(self, line):
        """
        history ID [END_DATE]

        Display history of transactions.

        If END_DATE is supplied, list all transactions until this date.
        """
        return self.show_history('iter_history', line)

    def complete_coming(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_account()

    @defaultcount(10)
    def do_coming(self, line):
        """
        coming ID [END_DATE]

        Display future transactions.

        If END_DATE is supplied, show all transactions until this date.
        """
        return self.show_history('iter_coming', line)

    def complete_transfer(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_account()
        if len(args) == 3:
            return self._complete_account(args[1])

    def do_add_recipient(self, line):
        """
        add_recipient iban label

        Add a recipient to a backend.
        """
        if len(self.enabled_backends) > 1:
            print('Error: select a single backend to add a recipient (Hint: try the command "backends only")', file=self.stderr)
            return 1
        iban, label, origin_account_id = self.parse_command_args(line, 3, 2)
        recipient = Recipient()
        recipient.iban = iban
        recipient.label = label
        recipient.origin_account_id = origin_account_id
        next(iter(self.do('add_recipient', recipient)))

    def do_recipients(self, line):
        """
        recipients ACCOUNT

        List recipients of ACCOUNT
        """
        id_from, = self.parse_command_args(line, 1, 1)

        account = self.get_object(id_from, 'get_account', [])
        if not account:
            print('Error: account %s not found' % id_from, file=self.stderr)
            return 1

        self.objects = []

        self.start_format()
        for recipient in self.do('iter_transfer_recipients', account, backends=account.backend, caps=CapTransfer):
            self.cached_format(recipient)

    @contextmanager
    def use_cmd_formatter(self, cmd):
        self.set_formatter(self.commands_formatters.get(cmd, self.DEFAULT_FORMATTER))
        try:
            yield
        finally:
            self.flush()

    def _build_transfer(self, line):
        if self.interactive:
            id_from, id_to, amount, reason, exec_date = self.parse_command_args(line, 5, 0)
        else:
            id_from, id_to, amount, reason, exec_date = self.parse_command_args(line, 5, 3)

        missing = not bool(id_from and id_to and amount)

        if id_from:
            account = self.get_object(id_from, 'get_account', [])
            id_from = account.id
            if not account:
                print('Error: account %s not found' % id_from, file=self.stderr)
                return
        else:
            with self.use_cmd_formatter('list'):
                self.do_ls('')
            id_from = self.ask('Transfer money from account', default='')
            if not id_from:
                return
            id_from, backend = self.parse_id(id_from)

            account = find_object(self.objects, fullid='%s@%s' % (id_from, backend))
            if not account:
                return
            id_from = account.id

        if id_to:
            id_to, backend_name_to = self.parse_id(id_to)
            if account.backend != backend_name_to:
                print("Transfer between different backends is not implemented", file=self.stderr)
                return
            rcpts = self.do('iter_transfer_recipients', id_from, backends=account.backend)
            rcpt = find_object(rcpts, id=id_to)
        else:
            with self.use_cmd_formatter('recipients'):
                self.do_recipients(account.fullid)
            id_to = self.ask('Transfer money to recipient', default='')
            if not id_to:
                return
            id_to, backend = self.parse_id(id_to)

            rcpt = find_object(self.objects, fullid='%s@%s' % (id_to, backend))
            if not rcpt:
                return

        if not amount:
            amount = self.ask('Amount to transfer', default='', regexp=r'\d+(?:\.\d*)?')
        try:
            amount = Decimal(amount)
        except (TypeError, ValueError, InvalidOperation):
            print('Error: please give a decimal amount to transfer', file=self.stderr)
            return
        if amount <= 0:
            print('Error: transfer amount must be strictly positive', file=self.stderr)
            return

        if missing:
            reason = self.ask('Label of the transfer (seen by the recipient)', default='')
            exec_date = self.ask('Execution date of the transfer (YYYY-MM-DD format, empty for today)', default='')

        today = datetime.date.today()
        if exec_date:
            try:
                exec_date = datetime.datetime.strptime(exec_date, '%Y-%m-%d').date()
            except ValueError:
                print('Error: execution date must be valid and in YYYY-MM-DD format', file=self.stderr)
                return
            if exec_date < today:
                print('Error: execution date cannot be in the past', file=self.stderr)
                return
        else:
            exec_date = today

        transfer = Transfer()
        transfer.backend = account.backend
        transfer.account_id = account.id
        transfer.account_label = account.label
        transfer.account_iban = account.iban
        transfer.recipient_id = id_to
        if rcpt:
            # Try to find the recipient label. It can be missing from
            # recipients list, for example for banks which allow transfers to
            # arbitrary recipients.
            transfer.recipient_label = rcpt.label
            transfer.recipient_iban = rcpt.iban
        transfer.amount = amount
        transfer.label = reason or ''
        transfer.exec_date = exec_date

        return transfer

    def do_transfer(self, line):
        """
        transfer [ACCOUNT RECIPIENT AMOUNT [LABEL [EXEC_DATE]]]

        Make a transfer beetwen two accounts
        - ACCOUNT    the source account
        - RECIPIENT  the recipient
        - AMOUNT     amount to transfer
        - LABEL      label of transfer
        - EXEC_DATE  date when to execute the transfer
        """

        transfer = self._build_transfer(line)
        if transfer is None:
            return 1

        if self.interactive:
            with self.use_cmd_formatter('transfer'):
                self.start_format()
                self.cached_format(transfer)

            if not self.ask('Are you sure to do this transfer?', default=True):
                return

        # only keep basic fields because most modules don't handle others
        transfer.account_label = None
        transfer.account_iban = None
        transfer.recipient_label = None
        transfer.recipient_iban = None

        self.start_format()
        next(iter(self.do('transfer', transfer, backends=transfer.backend)))

    def complete_transfer_history(self, text, line, *ignored):
        return self.complete_history(self, text, line, *ignored)

    @defaultcount(10)
    def do_transfer_history(self, line):
        """
        transfer_history [ACCOUNT_ID]

        Display history of transfer transactions.
        """
        id, = self.parse_command_args(line, 1, 0)

        account = None
        backends = None
        if id:
            account = self.get_object(id, 'get_account', [])
            if not account:
                print('Error: account "%s" not found (Hint: try the command "list")' % id, file=self.stderr)
                return 2
            backends = account.backend

        self.start_format()
        for tr in self.do('iter_transfers', account, backends=backends):
            self.format(tr)

    def show_wealth(self, command, id):
        account = self.get_object(id, 'get_account', [])
        if not account:
            print('Error: account "%s" not found (Hint: try the command "list")' % id, file=self.stderr)
            return 2

        caps = {
            'iter_investment': CapBankWealth,
            'iter_pocket': CapBankWealth,
            'iter_market_orders': CapBankWealth,
        }

        self.start_format()
        for el in self.do(command, account, backends=account.backend, caps=caps[command]):
            self.format(el)

    def do_investment(self, id):
        """
        investment ID

        Display investments of an account.
        """
        self.show_wealth('iter_investment', id)

    def do_pocket(self, id):
        """
        pocket ID

        Display pockets of an account.
        """
        self.show_wealth('iter_pocket', id)

    @defaultcount(10)
    def do_market_order(self, id):
        """
        market_order ID

        Display market orders of an account.
        """
        self.show_wealth('iter_market_orders', id)

    def do_profile(self, line):
        """
        profile

        Display detailed information about person or company.
        """
        self.start_format()
        for profile in self.do('get_profile', caps=CapProfile):
            self.format(profile)

    def do_emitters(self, line):
        """
        emitters

        Display transfer emitter account.
        """
        self.objects = []
        self.start_format()
        for emitter in self.do('iter_emitters', backends=list(self.enabled_backends), caps=CapTransfer):
            self.cached_format(emitter)

    def do_convert_currency(self, line):
        """
        convert_currency FROM_CURRENCY TO_CURRENCY [AMOUNT]

        Convert an amount from a currency to another
        """

        currency_from, currency_to, amount = self.parse_command_args(line, 3, 2)
        if amount is None:
            amount = 1
        amount = Decimal(amount)

        for rate in self.do("get_rate", currency_from, currency_to):
            if rate is not None:
                converted_amount = rate.convert(amount)
                print("%s %s is equal to %s %s" % (amount, rate.currency_from, converted_amount, rate.currency_to))
                break

    def main(self, argv):
        self.load_config()
        return super(Appbank, self).main(argv)



# Copyright(C) 2010-2012  Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from woob.capabilities.torrent import CapTorrent, MagnetOnly
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.core import CallErrors
from woob.capabilities.base import NotAvailable, NotLoaded, empty


__all__ = ['AppTorrent']


def sizeof_fmt(num):
    for x in ['bytes', 'KB', 'MB', 'GB', 'TB']:
        if num < 1024.0:
            return "%-4.1f%s" % (num, x)
        num /= 1024.0


class TorrentInfoFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'size', 'seeders', 'leechers', 'url', 'files', 'description')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.name, self.NC)
        result += 'ID: %s\n' % obj.fullid
        if obj.size != NotAvailable and obj.size != NotLoaded:
            result += 'Size: %s\n' % sizeof_fmt(obj.size)
        result += 'Seeders: %s\n' % obj.seeders
        result += 'Leechers: %s\n' % obj.leechers
        result += 'URL: %s\n' % obj.url
        if hasattr(obj, 'magnet') and obj.magnet:
            result += 'Magnet URL: %s\n' % obj.magnet
        if obj.files:
            result += '\n%sFiles%s\n' % (self.BOLD, self.NC)
            for f in obj.files:
                result += ' * %s\n' % f
        result += '\n%sDescription%s\n' % (self.BOLD, self.NC)
        result += '%s' % obj.description
        return result


class TorrentListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'name', 'size', 'seeders', 'leechers')

    def get_title(self, obj):
        return obj.name

    NB2COLOR = ((0, 'red', None),
                (1, 'blue', None),
                (5, 'green', None),
                (10, 'green', 'bold'),
               )

    def _get_color(self, nb):
        if empty(nb):
            return self.colored('N/A', 'red')

        for threshold, _color, _attr in self.NB2COLOR:
            if nb >= threshold:
                color = _color
                attr = _attr

        return self.colored('%3d' % nb, color, attr)

    def get_description(self, obj):
        size = self.colored('%10s' % sizeof_fmt(obj.size), 'magenta')
        return '%s   (Seed: %s / Leech: %s)' % (size,
                                                self._get_color(obj.seeders),
                                                self._get_color(obj.leechers))


class AppTorrent(ReplApplication):
    APPNAME = 'torrent'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2010-YEAR Romain Bignon'
    DESCRIPTION = "Console application allowing to search for torrents on various trackers " \
                  "and download .torrent files."
    SHORT_DESCRIPTION = "search and download torrents"
    CAPS = CapTorrent
    EXTRA_FORMATTERS = {'torrent_list': TorrentListFormatter,
                        'torrent_info': TorrentInfoFormatter,
                       }
    COMMANDS_FORMATTERS = {'search':    'torrent_list',
                           'info':      'torrent_info',
                          }

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, id):
        """
        info ID

        Get information about a torrent.
        """
        torrent = self.get_object(id, 'get_torrent', ('id', 'name', 'size', 'seeders', 'leechers', 'url', 'files', 'description'))
        if not torrent:
            print('Torrent not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(torrent)

    def complete_getfile(self, text, line, *ignored):
        args = line.split(' ', 2)
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_getfile(self, line):
        """
        getfile ID [FILENAME]

        Get the .torrent file.
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.
        """
        id, dest = self.parse_command_args(line, 2, 1)

        torrent = self.get_object(id, 'get_torrent', ('description', 'files'))
        if not torrent:
            print('Torrent not found: %s' % id, file=self.stderr)
            return 3

        dest = self.obj_to_filename(torrent, dest, '{id}-{name}.torrent')

        try:
            for buf in self.do('get_torrent_file', torrent.id, backends=torrent.backend):
                if buf:
                    if dest == '-':
                        print(buf)
                    else:
                        try:
                            with open(dest, 'w') as f:
                                f.write(buf)
                        except IOError as e:
                            print('Unable to write .torrent in "%s": %s' % (dest, e), file=self.stderr)
                            return 1
                    return
        except CallErrors as errors:
            for backend, error, backtrace in errors:
                if isinstance(error, MagnetOnly):
                    print('Error(%s): No direct URL available, '
                          'please provide this magnet URL '
                          'to your client:\n%s' % (backend, error.magnet), file=self.stderr)
                    return 4
                else:
                    self.bcall_error_handler(backend, error, backtrace)

        print('Torrent "%s" not found' % id, file=self.stderr)
        return 3

    @defaultcount(10)
    def do_search(self, pattern):
        """
        search [PATTERN]

        Search torrents.
        """
        self.change_path(['search'])
        if not pattern:
            pattern = None

        self.start_format(pattern=pattern)
        for torrent in self.do('iter_torrents', pattern=pattern):
            self.cached_format(torrent)



# Copyright(C) 2011  Pierre Mazière
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

import hashlib
import tempfile
from typing import IO, ClassVar, TYPE_CHECKING

if TYPE_CHECKING:
    from woob.browser import Browser


try:
    from PIL import Image
except ImportError:
    raise ImportError('Please install python-imaging')


class VirtKeyboardError(Exception):
    pass


class VirtKeyboard:
    """
    Handle a virtual keyboard.
    """

    margin = None
    """
    Margin used by :meth:`get_symbol_coords` to reduce size
    of each "key" of the virtual keyboard. This attribute is always
    converted to a 4-tuple, and has the same semantic as the CSS
    ``margin`` property (top, right, bottom, right), in pixels.
    """

    codesep = ''
    """
    Output separator between code strings.

    See :func:`get_string_code`.
    """

    def __init__(self, file=None, coords=None, color=None, convert=None):
        # file: virtual keyboard image
        # coords: dictionary <value to return>:<tuple(x1,y1,x2,y2)>
        # color: color of the symbols in the image
        #        depending on the image, it can be a single value or a tuple
        # convert: if not None, convert image to this target type (for example 'RGB')

        if file is not None:
            assert color, 'No color provided !'
            self.load_image(file, color, convert)

        if type(self.margin) in (int, float):
            self.margin = (self.margin,) * 4
        elif self.margin is not None:
            if len(self.margin) == 2:
                self.margin = self.margin + self.margin
            elif len(self.margin) == 3:
                self.margin = self.margin + (self.margin[1],)
            assert len(self.margin) == 4

        if coords is not None:
            self.load_symbols(coords)

    def load_image(self, file, color, convert=None):
        self.image = Image.open(file)

        if convert is not None:
            self.image = self.image.convert(convert)

        self.bands = self.image.getbands()
        if isinstance(color, int) and not isinstance(self.bands, str) and len(self.bands) != 1:
            raise VirtKeyboardError("Color requires %i component but only 1 is provided"
                                    % len(self.bands))
        if not isinstance(color, int) and len(color) != len(self.bands):
            raise VirtKeyboardError("Color requires %i components but %i are provided"
                                    % (len(self.bands), len(color)))
        self.color = color

        self.width, self.height = self.image.size
        self.pixar = self.image.load()

    def load_symbols(self, coords):
        self.coords = {}
        self.md5 = {}
        for i in coords:
            coord = self.get_symbol_coords(coords[i])
            if coord == (-1, -1, -1, -1):
                continue
            self.coords[i] = coord
            self.md5[i] = self.checksum(self.coords[i])

    def check_color(self, pixel):
        return pixel == self.color

    def get_symbol_coords(self, coords):
        """Return narrow coordinates around symbol."""
        (x1, y1, x2, y2) = coords
        if self.margin:
            top, right, bottom, left = self.margin
            x1, y1, x2, y2 = x1 + left, y1 + top, x2 - right, y2 - bottom

        newY1 = -1
        newY2 = -1
        for y in range(y1, min(y2 + 1, self.height)):
            empty_line = True
            for x in range(x1, min(x2 + 1, self.width)):
                if self.check_color(self.pixar[x, y]):
                    empty_line = False
                    if newY1 < 0:
                        newY1 = y
                    break
            if newY1 >= 0 and not empty_line:
                newY2 = y
        newX1 = -1
        newX2 = -1
        for x in range(x1, min(x2 + 1, self.width)):
            empty_column = True
            for y in range(y1, min(y2 + 1, self.height)):
                if self.check_color(self.pixar[x, y]):
                    empty_column = False
                    if newX1 < 0:
                        newX1 = x
                    break
            if newX1 >= 0 and not empty_column:
                newX2 = x
        return (newX1, newY1, newX2, newY2)

    def checksum(self, coords):
        (x1, y1, x2, y2) = coords
        s = b''
        for y in range(y1, min(y2 + 1, self.height)):
            for x in range(x1, min(x2 + 1, self.width)):
                if self.check_color(self.pixar[x, y]):
                    s += b"."
                else:
                    s += b" "
        return hashlib.md5(s).hexdigest()

    def get_symbol_code(self, all_known_md5_for_symbol):
        if isinstance(all_known_md5_for_symbol, str):
            all_known_md5_for_symbol = [all_known_md5_for_symbol]

        current_md5_in_keyboard = self.md5

        for known_md5 in all_known_md5_for_symbol:
            for code, cur_md5 in current_md5_in_keyboard.items():
                if known_md5 == cur_md5:
                    return code
        raise VirtKeyboardError(
                'Code not found for these hashes "%s".'
                % all_known_md5_for_symbol)

    def get_string_code(self, string):
        return self.codesep.join(self.get_symbol_code(self.symbols[c]) for c in string)

    def check_symbols(self, symbols, dirname):
        # symbols: dictionary <symbol>:<md5 value>
        for s in symbols:
            try:
                self.get_symbol_code(symbols[s])
            except VirtKeyboardError:
                if dirname is None:
                    dirname = tempfile.mkdtemp(prefix='woob_session_')
                self.generate_MD5(dirname)
                raise VirtKeyboardError("Symbol '%s' not found; all symbol hashes are available in %s"
                                        % (s, dirname))

    def generate_MD5(self, dir):
        for i in self.coords:
            width = self.coords[i][2] - self.coords[i][0] + 1
            height = self.coords[i][3] - self.coords[i][1] + 1
            img = Image.new(''.join(self.bands), (width, height))
            matrix = img.load()
            for y in range(height):
                for x in range(width):
                    matrix[x, y] = self.pixar[self.coords[i][0] + x, self.coords[i][1] + y]
            img.save(dir + "/" + self.md5[i] + ".png")
        self.image.save(dir + "/image.png")


class MappedVirtKeyboard(VirtKeyboard):
    def __init__(self, file, document, img_element, color, map_attr="onclick", convert=None):
        map_id = img_element.attrib.get("usemap")[1:]
        map = document.find('//map[@id="%s"]' % map_id)
        if map is None:
            map = document.find('//map[@name="%s"]' % map_id)

        coords = {}
        for area in map.iter("area"):
            code = area.attrib.get(map_attr)
            area_coords = []
            for coord in area.attrib.get("coords").split(' ')[0].split(','):
                area_coords.append(int(coord))
            coords[code] = tuple(area_coords)

        super(MappedVirtKeyboard, self).__init__(file, coords, color, convert)


class GridVirtKeyboard(VirtKeyboard):
    """
    Make a virtual keyboard where "keys" are distributed on a grid.
    Example here: https://www.e-sgbl.com/portalserver/sgbl-web/login

    :param symbols: Sequence of symbols, ordered in the grid from left to
        right and up to down
    :type symbols: iterable
    :param cols: Column count of the grid
    :type cols: int
    :param rows: Row count of the grid
    :type rows: int
    :param image: File-like object to be used as data source
    :type image: file
    :param color: Color of the meaningful pixels
    :type color: 3-tuple
    :param convert: Mode to which convert color of pixels, see
        :meth:`Image.Image.convert` for more information
    """
    symbols = {}
    """Assocation table between symbols and md5s"""

    def __init__(self, symbols, cols, rows, image, color, convert=None):
        self.load_image(image, color, convert)

        tileW = self.width / cols
        tileH = self.height / rows
        positions = ((s, i * tileW % self.width, i // cols * tileH)
                     for i, s in enumerate(symbols))
        coords = dict((s, tuple(map(int, (x, y, x + tileW, y + tileH))))
                      for (s, x, y) in positions)

        super(GridVirtKeyboard, self).__init__()

        self.load_symbols(coords)


class SplitKeyboard:
    """Virtual keyboard for when the chars are in individual images, not a single grid"""

    char_to_hash = None
    """dict mapping password characters to image hashes"""

    codesep = ''
    """Output separator between symbols"""

    def __init__(self, code_to_filedata):
        """Create a SplitKeyboard

        :param code_to_filedata: dict mapping site codes to images data
        :type code_to_filedata: dict[str, str]
        """

        hash_to_code = {
            self.checksum(data): code for code, data in code_to_filedata.items()
        }

        self.char_to_code = {}
        for char, hashes in self.char_to_hash.items():
            if isinstance(hashes, str):
                hashes = (hashes,)

            for hash in hash_to_code:
                if hash in hashes:
                    self.char_to_code[char] = hash_to_code.pop(hash)
                    break
            else:
                path = tempfile.mkdtemp(prefix='woob_session_')
                self.dump(code_to_filedata.values(), path)
                raise VirtKeyboardError("Symbol '%s' not found; all symbol hashes are available in %s" % (char, path))

    def checksum(self, buffer):
        return hashlib.md5(self.convert(buffer)).hexdigest()

    def dump(self, files, path):
        for dat in files:
            md5 = hashlib.md5(dat).hexdigest()
            with open('%s/%s.png' % (path, md5), 'wb') as fd:
                fd.write(dat)

    def get_string_code(self, password):
        symbols = []
        for c in password:
            symbols.append(self.char_to_code[c])
        return self.codesep.join(symbols)

    def convert(self, buffer):
        return buffer

    @classmethod
    def create_from_url(cls, browser, code_to_url):
        code_to_file = {
            code: browser.open(url).content for code, url in code_to_url
        }
        return cls(code_to_file)


class Tile:
    """Tile of a image grid for SimpleVirtualKeyboard"""
    def __init__(self, matching_symbol, coords, image=None, md5=None):
        self.matching_symbol = matching_symbol
        self.coords = coords
        self.image = image
        self.md5 = md5


class SimpleVirtualKeyboard:
    """Handle a virtual keyboard where "keys" are distributed on a simple grid.

    :param cols: Column count of the grid
    :param rows: Row count of the grid
    :param file: File-like object to be used as data source
    :param convert: Mode to which convert color of pixels, see
        :meth:`Image.Image.convert` for more information
    :param matching_symbols: symbol that match all case of image grid from left to right and top
        to down, European reading way.
    :param matching_symbols_coords: dict mapping matching website symbols to their image coords
        (x0, y0, x1, y1) on grid image from left to right and top to
        down, European reading way. It's not symbols in the image.
    :param browser: Browser of woob session.
        Allow to dump tiles files in same directory than session folder
    """

    codesep: ClassVar[str] = ''
    """Output separator between matching symbols"""

    margin: ClassVar[tuple[int, int, int, int] | tuple[int, int] | int | None] = None
    """
    4-tuple(int), same as HTML margin: (top, right, bottom, left).
    or 2-tuple(int), (top = bottom, right = left),
    or int, top = right = bottom = left
    """

    tile_margin: ClassVar[tuple[int, int, int, int] | tuple[int, int] | int | None] = None
    """
    4-tuple(int), same as HTML margin: (top, right, bottom, left).
    or 2-tuple(int), (top = bottom, right = left),
    or int, top = right = bottom = left
    """

    symbols: ClassVar[dict[str, str | tuple[str, ...]]] = None
    """
    Association table between image symbols and md5s
    """

    convert: ClassVar[str | None] = None
    """
    Mode to which convert color of pixels, see
    :meth:`Image.Image.convert` for more information
    """

    tile_klass = Tile

    def __init__(
        self,
        file: IO,
        cols: int,
        rows: int,
        matching_symbols: list[str] | None = None,
        matching_symbols_coords: dict[str, tuple[int, int, int, int]] | None = None,
        browser: Browser | None = None
    ):
        self.cols = cols
        self.rows = rows

        # Needed even if init is overwrite
        self.path = self.build_path(browser)

        # Get self.image
        self.load_image(file, self.margin, self.convert)

        # Get self.tiles
        self.get_tiles(
            matching_symbols=matching_symbols,
            matching_symbols_coords=matching_symbols_coords
        )

        # Tiles processing
        self.cut_tiles(self.tile_margin)
        self.hash_md5_tiles()

    def build_path(self, browser=None):
        if browser and browser.responses_dirname:
            return browser.responses_dirname
        else:
            return tempfile.mkdtemp(prefix='woob_session_')

    def load_image(self, file, margin=None, convert=None):
        self.image = Image.open(file)
        # Resize image if margin is given
        if margin:
            self.image = self.cut_margin(self.image, margin)
        if convert:
            self.image = self.image.convert(convert)
        # Give possibility to alter image before get tiles, overwrite :func:`alter_image`.
        self.alter_image()
        self.width, self.height = self.image.size

    def alter_image(self):
        pass

    def cut_margin(self, image, margin):
        width, height = image.size

        # Verify the magin value format
        if type(margin) is int:
            margin = (margin, margin, margin, margin)
        elif len(margin) == 2:
            margin = (margin[0], margin[1], margin[0], margin[1])
        elif len(margin) == 4:
            margin = margin
        else:
            assert (len(margin) == 3) & (len(margin) > 4), \
                "Margin format is wrong."

        assert ((margin[0] + margin[2]) < height) & ((margin[1] + margin[3]) < width), \
            "Margin is too high, there is not enough pixel to cut."

        image = image.crop((
            0 + margin[3],
            0 + margin[0],
            width - margin[1],
            height - margin[2]
        ))
        return image

    def get_tiles(self, matching_symbols=None, matching_symbols_coords=None):
        self.tiles = []

        # Tiles coords are given
        if matching_symbols_coords:
            for matching_symbol in matching_symbols_coords:
                self.tiles.append(
                    self.tile_klass(
                        matching_symbol=matching_symbol,
                        coords=matching_symbols_coords[matching_symbol]
                    )
                )
            return

        assert (not self.width%self.cols) & (not self.height%self.rows), \
            "Image width and height are not multiple of cols and rows. Please resize image with attribute `margin`."

        # Tiles coords aren't given, calculate them
        self.tileW = self.width // self.cols
        self.tileH = self.height // self.rows

        # Matching symbols aren't given, default value is range(columns*rows)
        if not matching_symbols:
            matching_symbols = ['%s' % i for i in range(self.cols*self.rows)]

        assert len(matching_symbols) == (self.cols*self.rows), \
            "Number of website matching symbols is not equal to the number of cases on the image."

        # Calculate tiles coords for each matching symbol from 1-dimension to 2-dimensions
        for index, matching_symbol in enumerate(matching_symbols):
            coords = self.get_tile_coords_in_grid(index)
            self.tiles.append(self.tile_klass(matching_symbol=matching_symbol, coords=coords))

    def get_tile_coords_in_grid(self, case_index):
        # Get the top left pixel coords of the tile
        x0 = (case_index % self.cols) * self.tileW
        y0 = (case_index // self.cols) * self.tileH

        # Get the bottom right coords of the tile
        x1 = x0 + self.tileW
        y1 = y0 + self.tileH

        coords = (x0, y0, x1, y1)
        return(coords)

    def cut_tiles(self, tile_margin=None):
        for tile in self.tiles:
            tile.image = self.image.crop(tile.coords)

        # Resize tile if margin is given
        if tile_margin:
            for tile in self.tiles:
                tile.image = self.cut_margin(tile.image, tile_margin)

    def hash_md5_tiles(self):
        for tile in self.tiles:
            tile.md5 = hashlib.md5(tile.image.tobytes()).hexdigest()

    def dump_tiles(self, path):
        for tile in self.tiles:
            tile.image.save('{}/{}.png'.format(path, tile.md5))

    def get_string_code(self, password):
        word = []

        for digit in password:
            for tile in self.tiles:
                if tile.md5 in self.symbols[digit]:
                    word.append(tile.matching_symbol)
                    break
            else:
                # Dump file only if the symbol is not found
                self.dump_tiles(self.path)
                raise VirtKeyboardError(
                    "Symbol '%s' not found; all symbol hashes are available in %s" % (digit, self.path)
                )
        return self.codesep.join(word)



# Copyright(C) 2013 Julien Veyssier
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import codecs

from woob.capabilities.recipe import CapRecipe
from woob.capabilities.base import empty
from woob.tools.application.repl import ReplApplication, defaultcount
from woob.tools.application.formatters.iformatter import IFormatter, PrettyFormatter
from woob.tools.capabilities.recipe import recipe_to_krecipes_xml

__all__ = ['AppRecipes']


class RecipeInfoFormatter(IFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'preparation_time', 'ingredients', 'instructions')

    def format_obj(self, obj, alias):
        result = '%s%s%s\n' % (self.BOLD, obj.title, self.NC)
        result += 'ID: %s\n' % obj.fullid
        if not empty(obj.author):
            result += 'Author: %s\n' % obj.author
        if not empty(obj.preparation_time):
            result += 'Preparation time: %smin\n' % obj.preparation_time
        if not empty(obj.cooking_time):
            result += 'Cooking time: %smin\n' % obj.cooking_time
        if not empty(obj.nb_person):
            nbstr = '-'.join(str(num) for num in obj.nb_person)
            result += 'Amount of people: %s\n' % nbstr
        result += '\n%sIngredients%s\n' % (self.BOLD, self.NC)
        for i in obj.ingredients:
            result += '  * %s\n' % i
        result += '\n%sInstructions%s\n' % (self.BOLD, self.NC)
        result += '%s\n' % obj.instructions
        if not empty(obj.comments):
            result += '\n%sComments%s\n' % (self.BOLD, self.NC)
            for c in obj.comments:
                result += '  * %s\n' % c
        return result


class RecipeListFormatter(PrettyFormatter):
    MANDATORY_FIELDS = ('id', 'title', 'short_description', 'preparation_time')

    def get_title(self, obj):
        return obj.title

    def get_description(self, obj):
        result = ''
        if not empty(obj.preparation_time):
            result += 'prep time: %smin' % obj.preparation_time
        if not empty(obj.short_description):
            result += 'description: %s\n' % obj.short_description
        return result.strip()


class AppRecipes(ReplApplication):
    APPNAME = 'recipes'
    VERSION = '3.7'
    COPYRIGHT = 'Copyright(C) 2013-YEAR Julien Veyssier'
    DESCRIPTION = "Console application allowing to search for recipes on various websites."
    SHORT_DESCRIPTION = "search and consult recipes"
    CAPS = CapRecipe
    EXTRA_FORMATTERS = {'recipe_list': RecipeListFormatter,
                        'recipe_info': RecipeInfoFormatter
                        }
    COMMANDS_FORMATTERS = {'search':    'recipe_list',
                           'info':      'recipe_info'
                           }

    def complete_info(self, text, line, *ignored):
        args = line.split(' ')
        if len(args) == 2:
            return self._complete_object()

    def do_info(self, id):
        """
        info ID

        Get information about a recipe.
        """
        recipe = self.get_object(id, 'get_recipe')
        if not recipe:
            print('Recipe not found: %s' % id, file=self.stderr)
            return 3

        self.start_format()
        self.format(recipe)

    def complete_export(self, text, line, *ignored):
        args = line.split(' ', 2)
        if len(args) == 2:
            return self._complete_object()
        elif len(args) >= 3:
            return self.path_completer(args[2])

    def do_export(self, line):
        """
        export ID [FILENAME]

        Export the recipe to a KRecipes XML file
        FILENAME is where to write the file. If FILENAME is '-',
        the file is written to stdout.
        """
        id, dest = self.parse_command_args(line, 2, 1)

        _id, backend_name = self.parse_id(id)

        if dest is None:
            dest = '%s.kreml' % _id

        recipe = self.get_object(id, 'get_recipe')

        if recipe:
            xmlstring = recipe_to_krecipes_xml(recipe, backend_name or None)
            if dest == '-':
                print(xmlstring)
            else:
                if not dest.endswith('.kreml'):
                    dest += '.kreml'
                try:
                    with codecs.open(dest, 'w', 'utf-8') as f:
                        f.write(xmlstring)
                except IOError as e:
                    print('Unable to write .kreml in "%s": %s' % (dest, e), file=self.stderr)
                    return 1
            return
        print('Recipe "%s" not found' % id, file=self.stderr)
        return 3

    @defaultcount(10)
    def do_search(self, pattern):
        """
        search [PATTERN]

        Search recipes.
        """
        self.change_path(['search'])
        self.start_format(pattern=pattern)
        for recipe in self.do('iter_recipes', pattern=pattern):
            self.cached_format(recipe)



# Copyright(C) 2010-2013 Christophe Benz, Romain Bignon, Julien Hebert
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from __future__ import annotations

from typing import Dict, Iterable, Type, Any, Tuple, TypeVar, overload

from collections import OrderedDict, deque
import warnings
import re
from decimal import Decimal
from copy import deepcopy, copy

from woob.tools.misc import to_unicode


__all__ = [
    'UserError', 'FieldNotFound', 'NotAvailable', 'FetchError', 'NotLoaded',
    'Capability', 'Field', 'IntField', 'DecimalField', 'FloatField',
    'StringField', 'BytesField', 'BoolField', 'Enum', 'EnumField', 'empty',
    'BaseObject', 'find_object', 'find_object_any_match', 'strict_find_object',
    'capability_to_string',
]


class EnumMeta(type):
    @classmethod
    def __prepare__(mcs, name, bases, **kwargs):
        # in python3.6, default namespace keeps declaration order
        # in python>=3 but <3.6, force ordered namespace
        # doesn't work in python2
        return OrderedDict()

    def __init__(cls, name, bases, attrs, *args, **kwargs):
        super(EnumMeta, cls).__init__(name, bases, attrs, *args, **kwargs)
        attrs = [(k, v) for k, v in attrs.items() if not callable(v) and not k.startswith('__')]
        cls.__members__ = OrderedDict(attrs)

    def __setattr__(cls, name, value):
        super(EnumMeta, cls).__setattr__(name, value)
        if not callable(value) and not name.startswith('__'):
            cls.__members__[name] = value

    def __call__(cls, *args, **kwargs):
        raise ValueError("Enum type can't be instanciated")

    @property
    def _items(cls):
        return cls.__members__.items()

    @property
    def _keys(cls):
        return cls.__members__.keys()

    @property
    def _values(cls):
        return cls.__members__.values()

    @property
    def _types(cls):
        return set(map(type, cls._values))

    def __iter__(cls):
        return iter(cls.__members__.values())

    def __len__(cls):
        return len(cls.__members__)

    def __contains__(cls, value):
        return value in cls.__members__.values()

    def __getitem__(cls, k):
        return cls.__members__[k]


class Enum(metaclass=EnumMeta):
    pass


def empty(value: Any) -> bool:
    """
    Checks if a value is empty (None, NotLoaded or NotAvailable).

    :rtype: :class:`bool`
    """
    return value is None or isinstance(value, EmptyType)


T = TypeVar('T')

@overload
def find_object(
    mylist: Iterable[T],
    error: None = None,
    **kwargs
) -> T | None:
    ...

@overload
def find_object(
    mylist: Iterable[T],
    error: Type[Exception],
    **kwargs
) -> T:
    ...

def find_object(
    mylist,
    error = None,
    **kwargs
):
    """
    Very simple tools to return an object with the matching parameters in
    kwargs.
    """
    for a in mylist:
        for key, value in kwargs.items():
            if getattr(a, key) != value:
                break
        else:
            return a

    if error is not None:
        raise error()
    return None


@overload
def find_object_any_match(
    objects: Iterable[T],
    key_value_pairs: Iterable[Tuple[str, Any]],
    error: None = None,
    with_priority: bool = True,
    ignore_empty: bool = True
) -> T | None:
    ...

@overload
def find_object_any_match(
    objects: Iterable[T],
    key_value_pairs: Iterable[Tuple[str, Any]],
    error: Type[Exception],
    with_priority: bool = True,
    ignore_empty: bool = True
) -> T:
    ...

def find_object_any_match(
    objects: Iterable[T],
    key_value_pairs: Iterable[Tuple[str, Any]],
    error: Type[Exception] | None = None,
    with_priority: bool = True,
    ignore_empty: bool = True
) -> T | None:
    """
    Tool method that returns an object that match any parameter given in key_value_pairs.

    Compared with find_object, this method does not need all parameters to be matching.
    If no priority is set, the method will return the first successful match. With several key-values set,
    this may not be the object instance with the most key matches.

    :param objects: The list or iterator of objects that can match
    :type objectlist: list
    :param key_value_pairs: The key-values that the object should have to match. This is given as a list
        of pairs, where each pair represent a key-value, the first item being the key, and the second is the value.
    :type key_value_pairs: list
    :param error: The error to raise in case no object matches.
    :type error: Exception or None
    :param with_priority: If True, the key-value pairs order decides the priority of the matching.
        With priority enabled, an object matching with a more important key will be matched over another
        object matching with a less important keys. This means that all objects will be iterated in order
        to assure this priority, unless the most preferred key matches, in which case the iteration stops early.
        This feature is enabled by default. See the example sections for more details.
    :type with_priority: bool
    :param ignore_empty: If True, empty key-values (None, EmptyType) will not be matched. True by default.
    :type ignore_empty: bool

    :return: The matching object if any

    Examples:

    >>> class Point:
    ...     def __init__(self, x, y):
    ...         self.x = x
    ...         self.y = y
    ...     def __repr__(self):
    ...         return f'Point({self.x},{self.y})'
    >>> points = (Point(0,0), Point(1, 2), Point(None, 3))
    >>> find_object_any_match(points, (('x', 0), ('y', 1))) # Will match on x with first point
    Point(0,0)
    >>> find_object_any_match(points, (('x', 2), ('y', 2))) # Will match on y with second point
    Point(1,2)
    >>> find_object_any_match(points, (('x', 1), ('y', 0))) # Will match on x with second point because priority
    Point(1,2)
    >>> find_object_any_match(points, (('x', 1), ('y', 0)), with_priority=False) # Will return first successful match
    Point(0,0)
    >>> find_object_any_match(points, (('x', None),)) # second condition (None) is ignored by default
    None
    >>> find_object_any_match(points, (('x', None),), ignore_empty=False) # will match last point
    Point(None,3)
    >>> find_object_any_match(points, (('x', -1),), error=ValueError) # Will raise ValueError
    ValueError:
    """

    # Will be used if with_priority is true
    found = None
    found_index = None

    for o in objects:
        for index, (key, value) in enumerate(key_value_pairs):
            # There is no point trying to match as it will be not priorized
            if found_index and found_index <= index:
                break

            o_value = getattr(o, key)
            if (
                ignore_empty and (empty(value) or empty(o_value))
                or o_value != value
            ):
                continue

            # First key is a special case with priority, we can return early
            # if we find a match on the first key
            # Also we only keep more priorized matches
            if not with_priority or index == 0:
                return o

            found = o
            found_index = index

    if not found and error is not None:
        raise error()
    return found


def strict_find_object(mylist: Iterable[BaseObject], error=None, **kwargs) -> BaseObject | None:
    """
    Tools to return an object with the matching parameters in kwargs.
    Parameters with empty value are skipped
    """
    kwargs = {k: v for k, v in kwargs.items() if not empty(v)}
    if kwargs:
        return find_object(mylist, error=error, **kwargs)

    if error is not None:
        raise error()

    return None


class UserError(Exception):
    """
    Exception containing an error message for user.
    """


class FieldNotFound(Exception):
    """
    A field isn't found.

    :param obj: object
    :type obj: :class:`BaseObject`
    :param field: field not found
    :type field: :class:`Field`
    """

    def __init__(self, obj: BaseObject, field: str):
        super().__init__('Field "%s" not found for object %s' % (field, obj))


class ConversionWarning(UserWarning):
    """
    A field's type was changed when setting it.
    Ideally, the module should use the right type before setting it.
    """


class AttributeCreationWarning(UserWarning):
    """
    A non-field attribute has been created with a name not
    prefixed with a _.
    """


class EmptyType:
    """
    Parent class for NotAvailableType, NotLoadedType and FetchErrorType.
    """

    def __str__(self):
        return repr(self)

    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __nonzero__(self):
        return False

    __bool__ = __nonzero__


class NotAvailableType(EmptyType):
    """
    NotAvailable is a constant to use on non available fields.
    """

    def __repr__(self):
        return 'NotAvailable'

    def __str__(self):
        return 'Not available'


NotAvailable = NotAvailableType()


class NotLoadedType(EmptyType):
    """
    NotLoaded is a constant to use on not loaded fields.

    When you use :func:`woob.tools.backend.Module.fillobj` on a object based on :class:`BaseObject`,
    it will request all fields with this value.
    """

    def __repr__(self):
        return 'NotLoaded'

    def __str__(self):
        return 'Not loaded'


NotLoaded = NotLoadedType()


class FetchErrorType(EmptyType):
    """
    FetchError is a constant to use when parsing a non-mandatory field raises an exception.
    """

    def __repr__(self):
        return 'FetchError'

    def __str__(self):
        return 'Not mandatory'


FetchError = FetchErrorType()


class Capability:
    """
    This is the base class for all capabilities.

    A capability may define abstract methods (which raise :class:`NotImplementedError`)
    with an explicit docstring to tell backends how to implement them.

    Also, it may define some *objects*, using :class:`BaseObject`.
    """


class Field:
    """
    Field of a :class:`BaseObject` class.

    :param doc: docstring of the field
    :type doc: :class:`str`
    :param args: list of types accepted
    :param default: default value of this field. If not specified, :class:`NotLoaded` is used.
    """
    _creation_counter = 0

    def __init__(self, doc, *args, **kwargs):
        self.types = ()
        self.value = kwargs.get('default', NotLoaded)
        self.doc = doc
        self.mandatory = kwargs.get('mandatory', True)

        for arg in args:
            if isinstance(arg, type) or isinstance(arg, str):
                self.types += (arg,)
            else:
                raise TypeError('Arguments must be types or strings of type name')

        self._creation_counter = Field._creation_counter
        Field._creation_counter += 1

    def convert(self, value):
        """
        Convert value to the wanted one.
        """
        return value


class IntField(Field):
    """
    A field which accepts only :class:`int` types.
    """

    def __init__(self, doc, **kwargs):
        super().__init__(doc, int, **kwargs)

    def convert(self, value):
        return int(value)


class BoolField(Field):
    """
    A field which accepts only :class:`bool` type.
    """

    def __init__(self, doc, **kwargs):
        super().__init__(doc, bool, **kwargs)

    def convert(self, value):
        return bool(value)


class DecimalField(Field):
    """
    A field which accepts only :class:`decimal` type.
    """

    def __init__(self, doc, **kwargs):
        super().__init__(doc, Decimal, **kwargs)

    def convert(self, value):
        if isinstance(value, Decimal):
            return value
        return Decimal(value)


class FloatField(Field):
    """
    A field which accepts only :class:`float` type.
    """

    def __init__(self, doc, **kwargs):
        super(FloatField, self).__init__(doc, float, **kwargs)

    def convert(self, value):
        return float(value)


class StringField(Field):
    """
    A field which accepts only :class:`str` strings.
    """

    def __init__(self, doc, **kwargs):
        super(StringField, self).__init__(doc, str, **kwargs)

    def convert(self, value):
        return to_unicode(value)


class BytesField(Field):
    """
    A field which accepts only :class:`bytes` strings.
    """

    def __init__(self, doc, **kwargs):
        super(BytesField, self).__init__(doc, bytes, **kwargs)

    def convert(self, value):
        if isinstance(value, str):
            value = value.encode('utf-8')
        return bytes(value)


class EnumField(Field):
    def __init__(self, doc, enum, **kwargs):
        if not issubclass(enum, Enum):
            raise TypeError('invalid enum type: %r' % enum)
        super(EnumField, self).__init__(doc, *enum._types, **kwargs)
        self.enum = enum

    def convert(self, value):
        if value not in self.enum._values:
            raise ValueError('value %r does not belong to enum %s' % (value, self.enum))
        return value


class _BaseObjectMeta(type):
    def __new__(cls, name, bases, attrs):
        fields = [(field_name, attrs.pop(field_name)) for field_name, obj in list(attrs.items()) if isinstance(obj, Field)]
        fields.sort(key=lambda x: x[1]._creation_counter)

        new_class = super(_BaseObjectMeta, cls).__new__(cls, name, bases, attrs)
        if new_class._fields is None:
            new_class._fields = OrderedDict()
        else:
            new_class._fields = deepcopy(new_class._fields)
        new_class._fields.update(fields)

        if new_class.__doc__ is None:
            new_class.__doc__ = ''
        for name, field in new_class._fields.items():
            doc = '(%s) %s' % (', '.join([':class:`%s`' % v.__name__ if isinstance(v, type) else v for v in field.types]), field.doc)
            if field.value is not NotLoaded:
                doc += ' (default: %s)' % field.value
            new_class.__doc__ += '\n:var %s: %s' % (name, doc)
        return new_class


class BaseObject(metaclass=_BaseObjectMeta):
    """
    This is the base class for a capability object.

    A capability interface may specify to return several kind of objects, to formalise
    retrieved information from websites.

    As python is a flexible language where variables are not typed, we use a system to
    force backends to set wanted values on all fields. To do that, we use the :class:`Field`
    class and all derived ones.

    For example::

        class Transfer(BaseObject):
            " Transfer from an account to a recipient.  "

            amount =    DecimalField('Amount to transfer')
            date =      Field('Date of transfer', str, date, datetime)
            origin =    Field('Origin of transfer', int, str)
            recipient = Field('Recipient', int, str)

    The docstring is mandatory.
    """

    id: str | None = None
    backend: str | None = None
    _fields: Dict[str, Field] = {}

    # XXX remove it?
    url = StringField('url')

    def __init__(
        self,
        id: str = '',
        url: str | NotLoadedType | NotAvailableType = NotLoaded,
        backend=None
    ):
        self.id = id or ''
        self.backend = backend
        self._fields = deepcopy(self._fields)
        self.__setattr__('url', url)

    @property
    def fullid(self) -> str:
        """
        Full ID of the object, in form '**ID@backend**'.
        """
        return '%s@%s' % (self.id, self.backend)

    def __iscomplete__(self) -> bool:
        """
        Return True if the object is completed.

        It is useful when the object is a field of an other object which is
        going to be filled.

        The default behavior is to iter on fields (with iter_fields) and if
        a field is NotLoaded, return False.
        """
        for key, value in self.iter_fields():
            if value is NotLoaded:
                return False
        return True

    def copy(self) -> BaseObject:
        obj = copy(self)
        obj._fields = copy(self._fields)
        for k in obj._fields:
            obj._fields[k] = copy(obj._fields[k])
        return obj

    def __deepcopy__(self, memo) -> BaseObject:
        return self.copy()

    def set_empty_fields(self, value: Any, excepts=()):
        """
        Set the same value on all empty fields.

        :param value: value to set on all empty fields
        :param excepts: if specified, do not change fields listed
        """
        for key, old_value in self.iter_fields():
            if empty(old_value) and key not in excepts:
                setattr(self, key, value)

    def iter_fields(self) -> Iterable[Tuple[str, Any]]:
        """
        Iterate on the fields keys and values.

        Can be overloaded to iterate on other things.

        :rtype: iter[(key, value)]
        """

        if hasattr(self, 'id') and self.id is not None:
            yield 'id', self.id
        for name, field in self._fields.items():
            yield name, field.value

    def __eq__(self, obj) -> bool:
        if isinstance(obj, BaseObject):
            return self.backend == obj.backend and self.id == obj.id
        else:
            return False

    def __getattr__(self, name: str) -> Any:
        if self._fields is not None and name in self._fields:
            return self._fields[name].value
        else:
            raise AttributeError("'%s' object has no attribute '%s'" % (
                self.__class__.__name__, name))

    def __setattr__(self, name: str, value: Any):
        try:
            attr = (self._fields or {})[name]
        except KeyError:
            if name not in dir(self) and not name.startswith('_'):
                warnings.warn('Creating a non-field attribute %s. Please prefix it with _' % name,
                              AttributeCreationWarning, stacklevel=2)
            object.__setattr__(self, name, value)
        else:
            if not empty(value):
                try:
                    # Try to convert value to the wanted one.
                    nvalue = attr.convert(value)
                except (TypeError, ValueError, ArithmeticError):
                    # error during conversion, it will probably not
                    # match the wanted following types, so we'll
                    # raise ValueError.
                    pass
                else:
                    # If the value was converted
                    if nvalue is not value:
                        warnings.warn('Value %s was converted from %s to %s' %
                                      (name, type(value), type(nvalue)),
                                      ConversionWarning, stacklevel=2)
                    value = nvalue

            actual_types = _resolve_types(attr.types)

            if not isinstance(value, actual_types) and not empty(value):
                raise ValueError(
                    'Value for "%s" needs to be of type %r, not %r' % (
                        name, actual_types, type(value)))
            attr.value = value

    def __delattr__(self, name: str):
        try:
            self._fields.pop(name)
        except KeyError:
            object.__delattr__(self, name)

    def to_dict(self) -> Dict[str, Any]:
        def iter_decorate(d):
            for key, value in d:
                if key == 'id' and self.backend is not None:
                    value = self.fullid
                yield key, value

        fields_iterator = self.iter_fields()
        return OrderedDict(iter_decorate(fields_iterator))

    def __getstate__(self) -> Dict[str, Any]:
        d = self.to_dict()
        d.update((k, v) for k, v in self.__dict__.items() if k != '_fields')
        return d

    @classmethod
    def from_dict(cls, values: Dict[str, Any], backend: str | None = None):
        self = cls()

        for attr in values:
            setattr(self, attr, values[attr])

        return self

    def __setstate__(self, state: Dict[str, Any]):
        self._fields = deepcopy(self._fields)  # because yaml does not call __init__
        for k in state:
            setattr(self, k, state[k])

    def __dir__(self):
        return list(super(BaseObject, self).__dir__()) + list(self._fields.keys())


def _resolve_types(types):
    actual_types = ()
    for v in types:
        if isinstance(v, str):
            # the following is a (almost) copy/paste from
            # https://stackoverflow.com/questions/11775460/lexical-cast-from-string-to-type
            q = deque([object])
            while q:
                t = q.popleft()
                if t.__name__ == v:
                    actual_types += (t,)
                else:
                    try:
                        # keep looking!
                        q.extend(t.__subclasses__())
                    except TypeError:
                        # type.__subclasses__ needs an argument for
                        # whatever reason.
                        if t is type:
                            continue
                        else:
                            raise
        else:
            actual_types += (v,)

    return actual_types


class Currency:
    CURRENCIES = OrderedDict([
        ('EUR', ('€', 'EURO', 'EUROS')),
        ('CHF', ('CHF',)),
        ('USD', ('$', '$US')),
        ('GBP', ('£',)),
        ('LBP', ('ل.ل',)),
        ('AED', ('AED',)),
        ('XOF', ('XOF',)),
        ('RUB', ('руб',)),
        ('SGD', ('SGD',)),
        ('BRL', ('R$',)),
        ('MXN', ('$',)),
        ('JPY', ('¥',)),
        ('TRY', ('₺', 'TRY')),
        ('RON', ('lei',)),
        ('COP', ('$',)),
        ('NOK', ('kr',)),
        ('CNY', ('¥',)),
        ('RSD', ('din',)),
        ('ZAR', ('rand',)),
        ('MYR', ('RM',)),
        ('HUF', ('Ft',)),
        ('HKD', ('HK$',)),
        ('TWD', ('NT$',)),
        ('QAR', ('QR',)),
        ('MAD', ('MAD',)),
        ('ARS', ('ARS',)),
        ('AUD', ('AUD',)),
        ('CAD', ('CAD',)),
        ('NZD', ('NZD',)),
        ('BHD', ('BHD',)),
        ('SEK', ('SEK',)),
        ('DKK', ('DKK',)),
        ('LUF', ('LUF',)),
        ('KZT', ('KZT',)),
        ('PLN', ('PLN',)),
        ('ILS', ('ILS',)),
        ('THB', ('THB',)),
        ('INR', ('₹', 'INR')),
        ('PEN', ('S/',)),
        ('IDR', ('Rp',)),
        ('KWD', ('KD',)),
        ('KRW', ('₩',)),
        ('CZK', ('Kč',)),
        ('EGP', ('E£',)),
        ('ISK', ('Íkr', 'kr')),
        ('XPF', ('XPF',)),
        ('SAR', ('SAR',)),
        ('BGN', ('лв',)),
        ('HRK', ('HRK',  'kn')),
    ])

    EXTRACTOR = re.compile(r'[()\d\s,\.\-]', re.UNICODE)

    @classmethod
    def get_currency(cls: Type[Currency], text: str) -> str | None:
        """
        >>> Currency.get_currency(u'42')
        None
        >>> Currency.get_currency(u'42 €')
        u'EUR'
        >>> Currency.get_currency(u'$42')
        u'USD'
        >>> Currency.get_currency(u'42.000,00€')
        u'EUR'
        >>> Currency.get_currency(u'$42 USD')
        u'USD'
        >>> Currency.get_currency(u'%42 USD')
        u'USD'
        >>> Currency.get_currency(u'US1D')
        None
        """
        curtexts = cls.EXTRACTOR.sub(' ', text.upper()).split()

        for currency, symbols in cls.CURRENCIES.items():
            for curtext in curtexts:
                if curtext == currency:
                    return currency
                for symbol in symbols:
                    if curtext == symbol:
                        return currency
        return None

    @classmethod
    def currency2txt(cls, currency):
        _currency = cls.CURRENCIES.get(currency, ('',))
        return _currency[0]


def capability_to_string(capability_klass: Type[Capability]) -> str:
    m = re.match(r'^Cap(\w+)', capability_klass.__name__)
    assert m is not None

    return m.group(1).lower()


class DeprecatedFieldWarning(UserWarning):
    pass



# Copyright(C) 2012 Romain Bignon, Florent Fourcot
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

import warnings

from woob.exceptions import BrowserIncorrectPassword

from .account import CapCredentialsCheck
from .base import (
    BaseObject, BoolField, Currency, DecimalField, DeprecatedFieldWarning,
    Enum, Field, StringField, UserError, empty, find_object,
)
from .date import DateField
from .collection import CapCollection


__all__ = ['SubscriptionNotFound', 'DocumentNotFound', 'DocumentTypes', 'Detail', 'Document', 'Bill', 'Subscription', 'CapDocument']


class SubscriptionNotFound(UserError):
    """
    Raised when a subscription is not found.
    """

    def __init__(self, msg='Subscription not found'):
        super(SubscriptionNotFound, self).__init__(msg)


class DocumentNotFound(UserError):
    """
    Raised when a document is not found.
    """

    def __init__(self, msg='Document not found'):
        super(DocumentNotFound, self).__init__(msg)


class DocumentTypes:
    RIB = 'RIB'
    """French bank account identification document"""

    STATEMENT = 'statement'
    """Bank statement"""

    CONTRACT = 'contract'
    """Contract between organisation and subscriber"""

    CERTIFICATE = 'certificate'
    """Certificate from the organisation to the subscriber"""

    NOTICE = 'notice'
    """Notice from the organisation to the subscriber"""

    REPORT = 'report'
    """Informative report"""

    BILL = 'bill'
    """Bill"""

    KIID = 'kiid'
    """Key Investor Information Document"""

    IDENTITY = 'identity'
    """Identity document (e.g. national identity card, driving licence, etc...)"""

    PAYSLIP = 'payslip'
    """Payslip"""

    OTHER = 'other'


class Detail(BaseObject, Currency):
    """
    Detail of a subscription
    """
    label =     StringField('label of the detail line')
    infos =     StringField('information')
    datetime =  DateField('date information')
    price =     DecimalField('Total price, taxes included')
    vat =       DecimalField('Value added Tax')
    currency =  StringField('Currency', default=None)
    quantity =  DecimalField('Number of units consumed')
    unit =      StringField('Unit of the consumption')


class Document(BaseObject):
    """
    Document.
    """
    date =          DateField('The day the document has been sent to the subscriber')
    format =        StringField('file format of the document')
    label =         StringField('label of document')
    type =          StringField('type of document')
    transactions =  Field('List of transaction ID related to the document', list, default=[])
    has_file =      BoolField('Boolean to set if file is available', default=True)
    number =        StringField('Number of the document (if present and meaningful for user)')

    def __repr__(self):
        return '<%s id=%r label=%r date=%r>' % (type(self).__name__, self.id, self.label, self.date)


class Bill(Document, Currency):
    """
    Bill.
    """
    total_price =   DecimalField('Price to pay')
    currency =      StringField('Currency', default=None)
    vat =           DecimalField('VAT included in the price')
    pre_tax_price = DecimalField('Price without the VAT or other taxes')
    duedate =       DateField('The day the bill must be paid')
    startdate =     DateField('The first day the bill applies to')
    finishdate =    DateField('The last day the bill applies to')

    def __repr__(self):
        return '<%s id=%r label=%r date=%r total_price=%r>' % (
            type(self).__name__, self.id, self.label, self.date, self.total_price
        )

    # compatibility properties
    @property
    def price(self):
        warnings.warn(
            'Field "price" is deprecated, use "total_price" field instead.',
            DeprecatedFieldWarning, stacklevel=3,
        )

        if empty(self.total_price):
            return self.total_price
        return abs(self.total_price)

    @price.setter
    def price(self, value):
        warnings.warn(
            'Field "price" is deprecated, use "total_price" field instead.',
            DeprecatedFieldWarning, stacklevel=3,
        )
        if empty(value):
            self.total_price = value
            self._income = None
            return

        value = abs(value)
        if not self.income:
            self.total_price = value
        else:
            self.total_price = -value
        self._income = None

    @property
    def income(self):
        warnings.warn(
            'Field "income" is deprecated, use "total_price" field instead.',
            DeprecatedFieldWarning, stacklevel=3,
        )

        if empty(self.total_price):
            return self._income or False
        return self.total_price <= 0

    @income.setter
    def income(self, value):
        warnings.warn(
            'Field "income" is deprecated, use "total_price" field instead.',
            DeprecatedFieldWarning, stacklevel=3,
        )

        if empty(self.total_price):
            self._income = value
        else:
            # the price was set before income
            # we can avoid handling the obnoxious _income field
            income_sign = {True: -1, False: 1}
            self.total_price = abs(self.total_price) * income_sign[value]
            self._income = None

    def __init__(self, *args, **kwargs):
        super(Bill, self).__init__(*args, **kwargs)
        self.type = DocumentTypes.BILL
        self._income = None
        # _income shall be set to True or False only if income is set *before* price property
        # in all other cases, we can map income property depending on total_price


class Subscription(BaseObject):
    """
    Subscription to a service.
    """
    label =         StringField('label of subscription')
    subscriber =    StringField('Subscriber name or identifier (for companies)')
    validity =      DateField('End validity date of the subscription (if any)')
    renewdate =     DateField('Reset date of consumption, for time based subscription (monthly, yearly, etc)')
    is_fake =       BoolField("Set True if website doesn't provide a real subscription", default=False)

    def __repr__(self):
        return '<%s id=%r label=%r>' % (type(self).__name__, self.id, self.label)


class CapDocument(CapCollection, CapCredentialsCheck):
    accepted_document_types = ()
    """
    Tuple of document types handled by the module (:class:`DocumentTypes`)
    """

    def check_credentials(self) -> bool:
        """
        Check that the given credentials are correct by trying to login.

        The default implementation of this method check if the class using this capability
        has a browser, execute its do_login if it has one and then see if no error pertaining to the creds is raised.
        If any other unexpected error occurs, we don't know whether the creds are correct or not.
        """
        if getattr(self, 'BROWSER', None) is None:
            raise NotImplementedError()

        try:
            self.browser.do_login()
        except BrowserIncorrectPassword:
            return False

        return True

    def iter_subscription(self):
        """
        Iter subscriptions.

        :rtype: iter[:class:`Subscription`]
        """
        raise NotImplementedError()

    def get_subscription(self, _id):
        """
        Get a subscription.

        :param _id: ID of subscription
        :rtype: :class:`Subscription`
        :raises: :class:`SubscriptionNotFound`
        """
        return find_object(
            self.iter_subscription(),
            id=_id,
            error=SubscriptionNotFound,
        )

    def iter_documents_history(self, subscription):
        """
        Iter history of a subscription.

        :param subscription: subscription to get history
        :type subscription: :class:`Subscription`
        :rtype: iter[:class:`Detail`]
        """
        raise NotImplementedError()

    def iter_bills_history(self, subscription):
        """
        Iter history of a subscription.

        :param subscription: subscription to get history
        :type subscription: :class:`Subscription`
        :rtype: iter[:class:`Detail`]
        """
        return self.iter_documents_history(subscription)

    def get_document(self, id):
        """
        Get a document.

        :param id: ID of document
        :rtype: :class:`Document`
        :raises: :class:`DocumentNotFound`
        """
        raise NotImplementedError()

    def download_document(self, id):
        """
        Download a document.

        :param id: ID of document
        :rtype: bytes
        :raises: :class:`DocumentNotFound`
        """
        raise NotImplementedError()

    def download_document_pdf(self, id):
        """
        Download a document, convert it to PDF if it isn't the document format.

        :param id: ID of document
        :rtype: bytes
        :raises: :class:`DocumentNotFound`
        """
        if not isinstance(id, Document):
            id = self.get_document(id)

        if id.format == 'pdf':
            return self.download_document(id)
        else:
            raise NotImplementedError()

    def iter_documents(self, subscription):
        """
        Iter documents.

        :param subscription: subscription to get documents
        :type subscription: :class:`Subscription`
        :rtype: iter[:class:`Document`]
        """
        raise NotImplementedError()

    def iter_documents_by_types(self, subscription, accepted_types):
        """
        Iter documents with specific types.

        :param subscription: subscription to get documents
        :type subscription: :class:`Subscription`
        :param accepted_types: list of document types that should be returned
        :type accepted_types: [:class:`DocumentTypes`]
        :rtype: iter[:class:`Document`]
        """
        accepted_types = frozenset(accepted_types)
        for document in self.iter_documents(subscription):
            if document.type in accepted_types:
                yield document

    def iter_bills(self, subscription):
        """
        Iter bills.

        :param subscription: subscription to get bills
        :type subscription: :class:`Subscription`
        :rtype: iter[:class:`Bill`]
        """
        documents = self.iter_documents(subscription)
        return [doc for doc in documents if doc.type == "bill"]

    def get_details(self, subscription):
        """
        Get details of a subscription.

        :param subscription: subscription to get details
        :type subscription: :class:`Subscription`
        :rtype: iter[:class:`Detail`]
        """
        raise NotImplementedError()

    def get_balance(self, subscription):
        """
        Get the balance of a subscription.

        :param subscription: subscription to get balance
        :type subscription: :class:`Subscription`
        :rtype: class:`Detail`
        """
        raise NotImplementedError()

    def iter_resources(self, objs, split_path):
        """
        Iter resources. Will return :func:`iter_subscriptions`.
        """
        if Subscription in objs:
            self._restrict_level(split_path)
            return self.iter_subscription()


class CapDocumentMatching(CapDocument):
    """
    Capability for matching data between synchronizations.

    This is mostly useful for providers which have to compare states across time.
    For example, a provider has to compare subscriptions freshly returned to subscriptions
    returned in a previous sync.
    """

    def match_subscription(self, subscription, old_subscriptions):
        """Search a subscription in `old_subscriptions` corresponding to `subscription`.

        `old_subscriptions` is a list of subscriptions found in a previous
        synchronisation.
        However, they may not be the exact same objects but only reconstructed
        objects with the same data, although even it could be partial.
        For example, they may have been marshalled, sometimes loosely, thus some
        attributes may be missing (like `_private` attributes) or unset (some
        providers may choose not to even save all attributes).
        Also, `old_subscriptions` may not contain all subscriptions from previous state,
        but only subscriptions which have not been matched yet.

        :param subscription: fresh subscription to search for
        :type subscription: :class:`Subscription`
        :param old_subscriptions: candidates subscriptions from previous sync
        :type old_subscriptions: iter[:class:`Subscription`]
        :return: the corresponding subscription from `old_subscriptions`, or `None` if none matches
        :rtype: :class:`Subscription`
        """

        raise NotImplementedError()

    def match_document(self, document, old_documents):
        """
        :param document: fresh document to search for
        :type document: :class:`Bill`
        :param old_documents: candidates documents from previous sync
        :type old_documents: iter[:class:`Bill`]
        :return: the corresponding document from `old_documents`, or `None` if none matches
        :rtype: :class:`Bill`
        """

        raise NotImplementedError()


class DocumentCategory(Enum):
    """
    used for categories attribute in module.py
    use OTHER category sparingly,
    prefer define a new value if you really need a new one
    respect uppercase for enum and titled for string

    tag Module with it:
    example:
        document_categories = {DocumentCategory.ENERGY}
    """

    ADMINISTRATIVE = 'Administrative'
    BANK_INSURANCE = 'Bank & Insurance'
    BILLING = 'Billing'
    BUSINESS = 'Business'
    ENERGY = 'Energy Provider'
    FOOD = 'Food'
    INTERNET_TELEPHONY = 'Internet & Telephony'
    LEISURE = 'Leisure'
    OTHER = 'Other'
    REAL_ESTATE = 'Real estate'
    SAFE_DEPOSIT_BOX = 'Safe deposit box'
    """The Safe Deposit Box category concerns website whose purpose is to store
    official and sensitive documents such as proof of identity or wage slip,
    filed by companies and other legal entities."""
    SHOPPING = 'Shopping'
    SOFTWARE = 'Software'
    TAX = 'Tax'
    TOURISM = 'Tourism'
    TRANSPORT = 'Transport'
    WATER = 'Water Provider'



# Copyright(C) 2014 Oleg Plakhotniuk
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


from .base import BaseObject, StringField, DecimalField, UserError
from .date import DateField
from .collection import CapCollection


__all__ = ['OrderNotFound', 'Order', 'Payment', 'Item', 'CapShop']


class OrderNotFound(UserError):
    """
    Raised when an order is not found.
    """

    def __init__(self, msg='Order not found'):
        super(OrderNotFound, self).__init__(msg)


class Order(BaseObject):
    """
    Purchase order.
    """
    date     = DateField('Date when the order was placed')
    shipping = DecimalField('Shipping price')
    discount = DecimalField('Discounts')
    tax      = DecimalField('Tax')
    total    = DecimalField('Total')

    def __repr__(self):
        return "<Order id=%r date=%r>" % (self.id, self.date)


class Payment(BaseObject):
    """
    Payment for an order.
    """
    date   = DateField('The date when payment was applied')
    method = StringField('Payment method; e.g. "VISA 1234"')
    amount = DecimalField('Payment amount')

    def __repr__(self):
        return "<Payment date=%r method=%r amount=%r>" % \
            (self.date, self.method, self.amount)


class Item(BaseObject):
    """
    Purchased item within an order.
    """
    label = StringField('Item label')
    price = DecimalField('Item price')

    def __repr__(self):
        return "<Item label=%r price=%r>" % (self.label, self.price)


class CapShop(CapCollection):
    """
    Capability of online shops to see orders history.
    """

    def iter_resources(self, objs, split_path):
        """
        Iter resources.

        Default implementation of this method is to return on top-level
        all orders (by calling :func:`iter_accounts`).

        :param objs: type of objects to get
        :type objs: tuple[:class:`BaseObject`]
        :param split_path: path to discover
        :type split_path: :class:`list`
        :rtype: iter[:class:`BaseObject`]
        """
        if Order in objs:
            self._restrict_level(split_path)
            return self.iter_orders()

    def get_currency(self):
        """
        Get the currency this shop uses.

        :rtype: :class:`str`
        """
        raise NotImplementedError()

    def iter_orders(self):
        """
        Iter history of orders.

        :rtype: iter[:class:`Order`]
        """
        raise NotImplementedError()

    def get_order(self, id):
        """
        Get an order from its ID.

        :param id: ID of the order
        :type id: :class:`str`
        :rtype: :class:`Order`
        :raises: :class:`OrderNotFound`
        """
        raise NotImplementedError()

    def iter_payments(self, order):
        """
        Iter payments of a specific order.

        :param order: order to get payments
        :type order: :class:`Order`
        :rtype: iter[:class:`Payment`]
        :raises: :class:`OrderNotFound`
        """
        raise NotImplementedError()

    def iter_items(self, order):
        """
        Iter items of a specific order.

        :param order: order to get items
        :type order: :class:`Order`
        :rtype: iter[:class:`Item`]
        :raises: :class:`OrderNotFound`
        """
        raise NotImplementedError()



# Copyright(C) 2010-2011 Christophe Benz, Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.


__all__ = ['FormattersLoader', 'FormatterLoadError']


class FormatterLoadError(Exception):
    pass


class FormattersLoader:
    BUILTINS = ['htmltable', 'multiline', 'simple', 'table', 'csv', 'json', 'json_line']

    def __init__(self):
        self.formatters = {}

    def register_formatter(self, name, klass):
        self.formatters[name] = klass

    def get_available_formatters(self):
        l = set(self.formatters)
        l = l.union(self.BUILTINS)
        l = sorted(l)
        return l

    def build_formatter(self, name):
        if name not in self.formatters:
            try:
                self.formatters[name] = self.load_builtin_formatter(name)
            except ImportError as e:
                FormattersLoader.BUILTINS.remove(name)
                raise FormatterLoadError('Unable to load formatter "%s": %s' % (name, e))
        return self.formatters[name]()

    def load_builtin_formatter(self, name):
        if name not in self.BUILTINS:
            raise FormatterLoadError('Formatter "%s" does not exist' % name)

        if name == 'htmltable':
            from .table import HTMLTableFormatter
            return HTMLTableFormatter
        elif name == 'table':
            from .table import TableFormatter
            return TableFormatter
        elif name == 'simple':
            from .simple import SimpleFormatter
            return SimpleFormatter
        elif name == 'multiline':
            from .multiline import MultilineFormatter
            return MultilineFormatter
        elif name == 'csv':
            from .csv import CSVFormatter
            return CSVFormatter
        elif name == 'json':
            from .json import JsonFormatter
            return JsonFormatter
        elif name == 'json_line':
            from .json import JsonLineFormatter
            return JsonLineFormatter



# Copyright(C) 2010-2016 Romain Bignon
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from .transfer import (
    EmitterNumberType,
    Emitter,
    TransferFrequency,
    TransferDateType,
    TransferStatus,
    TransferError,
    TransferBankError,
    TransferTimeout,
    TransferInvalidEmitter,
    TransferInvalidRecipient,
    TransferInvalidLabel,
    TransferInvalidAmount,
    TransferInvalidCurrency,
    TransferInsufficientFunds,
    TransferInvalidDate,
    TransferInvalidOTP,
    TransferCancelledByUser,
    TransferNotFound,
    BeneficiaryType,
    RecipientNotFound,
    RecipientInvalidLabel,
    Recipient,
    Transfer,
    TransferStep,
    AddRecipientError,
    AddRecipientBankError,
    AddRecipientTimeout,
    AddRecipientStep,
    RecipientInvalidOTP,
    RecipientInvalidIban,
    CapTransfer,
    CapBankTransfer,
    CapBankTransferAddRecipient,
)
from .base import (
    AccountNotFound,
    AccountType,
    Currency,
    TransactionType,
    AccountOwnerType,
    Account,
    Loan,
    Transaction,
    AccountOwnership,
    NoAccountsException,
    CapBank,
    AccountSchemeName,
    TransactionCounterparty,
    PartyIdentity,
    AccountParty,
    AccountIdentification,
    PartyRole,
    CapAccountCheck,
    Balance,
    BalanceType,
    BankTransactionCode,
)
from .rate import Rate, CapCurrencyRate
from .wealth import (
    Investment,
    Per,
    CapBankWealth,
)


__all__ = [
    'EmitterNumberType',
    'Emitter',
    'TransferFrequency',
    'TransferDateType',
    'TransferStatus',
    'TransferError',
    'TransferBankError',
    'TransferTimeout',
    'TransferInvalidEmitter',
    'TransferInvalidRecipient',
    'TransferInvalidLabel',
    'TransferInvalidAmount',
    'TransferInvalidCurrency',
    'TransferInsufficientFunds',
    'TransferInvalidDate',
    'TransferInvalidOTP',
    'TransferCancelledByUser',
    'TransferNotFound',
    'BeneficiaryType',
    'RecipientNotFound',
    'RecipientInvalidLabel',
    'Recipient',
    'Transfer',
    'TransferStep',
    'AddRecipientError',
    'AddRecipientBankError',
    'AddRecipientTimeout',
    'AddRecipientStep',
    'RecipientInvalidOTP',
    'RecipientInvalidIban',
    'CapTransfer',
    'CapBankTransfer',
    'CapBankTransferAddRecipient',
    'AccountNotFound',
    'AccountType',
    'TransactionType',
    'AccountOwnerType',
    'Currency',
    'Account',
    'Loan',
    'Transaction',
    'AccountOwnership',
    'NoAccountsException',
    'CapBank',
    'Rate',
    'CapCurrencyRate',
    'Investment',
    'Per',
    'CapBankWealth',
    'AccountSchemeName',
    'TransactionCounterparty',
    'PartyIdentity',
    'AccountIdentification',
    'AccountParty',
    'PartyRole',
    'CapAccountCheck',
    'Balance',
    'BalanceType',
    'BankTransactionCode',
]



# Copyright(C) 2010-2017  Vincent Ardisson
#
# This file is part of woob.
#
# woob is free software: you can redistribute it and/or modify
# it under the terms of the GNU Lesser General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# woob is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public License
# along with woob. If not, see <http://www.gnu.org/licenses/>.

from functools import wraps


class SiteSwitch(Exception):
    """Exception to raise to switch to another Browser."""

    def __init__(self, name):
        """
        :param name: key of the `SwitchingBrowser.BROWSERS` dict to indicate
                     the new browser class to use
        :type name: str
        """
        super(SiteSwitch, self).__init__('Switching to site %s' % name)
        self.name = name


class SwitchingBrowser:
    """Proxy browser to use multiple (exclusive) browsers.

    When some sites have mutually exclusive sub-sites, it may be better to
    split a browser in multiple browsers. If it's not possible to know in
    advance what browser should be used, the SwitchingBrowser can help.

    Multiple browsers should be configured in the `BROWSERS` attribute as
    a dict. When first used, SwitchingBrowser will instanciate the browser
    class with the `'main'` key and proxy all method calls to it.
    If that browser raises :class:`SiteSwitch` exception, another browser
    (associated to the exception key parameter) will be instanciated and will
    be used to retry the call which failed.
    """

    BROWSERS = None

    """dict association keys to browser classes.

    It should contain a `'main'` key for the first browser class to use.
    """

    KEEP_SESSION = False

    """Whether to pass the :class:`requests.session.Session` between browsers.
    """

    KEEP_ATTRS = ()

    """Pass the values stored in __states__
    """

    def __init__(self, *args, **kwargs):
        super(SwitchingBrowser, self).__init__()
        self._browser_args = args
        self._browser_kwargs = kwargs
        self._browser = None

        self.set_browser('main')

    def set_browser(self, name):
        klass = self.BROWSERS[name]
        obj = klass(*self._browser_args, **self._browser_kwargs)
        if self._browser is not None:
            for attrname in self.KEEP_ATTRS:
                if hasattr(self._browser, attrname):
                    setattr(obj, attrname, getattr(self._browser, attrname))

            if self.KEEP_SESSION:
                obj.session = self._browser.session
            else:
                self._browser.session.close()

        self._browser = obj
        self._browser.logger.info('using %r browser', name)

    def __getattr__(self, attr):
        val = getattr(self._browser, attr)
        if not callable(val):
            return val

        @wraps(val)
        def wrapper(*args, **kwargs):
            try:
                return val(*args, **kwargs)
            except SiteSwitch as e:
                self.set_browser(e.name)
                val2 = getattr(self._browser, attr)
                return val2(*args, **kwargs)

        return wrapper


class SwitchingBrowserWithState(SwitchingBrowser):
    """Use state to transmit knowledge of last browser used during a previous sync to later start on the same browser"""

    def set_browser(self, name):
        super(SwitchingBrowserWithState, self).set_browser(name)
        self.last_browser = name

    def load_state(self, state):
        """Get the last used browser from the state, if any"""

        self.set_browser(name=state.get('last_browser', 'main'))
        self._browser.load_state(state)

    def dump_state(self):
        """Store the last used browser in the state"""

        ret = self._browser.dump_state()
        ret['last_browser'] = self.last_browser
        return ret
